import collections
import logging
import math
import os
import re
import shutil
import sys
from itertools import product

import numpy as np
from castepxbin.pdos import compute_pdos
from monty.io import zopen
from pymatgen.core.lattice import Lattice
from pymatgen.core.structure import Structure
from pymatgen.electronic_structure.bandstructure import BandStructureSymmLine
from pymatgen.electronic_structure.core import Spin
from pymatgen.electronic_structure.dos import CompleteDos, Dos
from pymatgen.phonon.bandstructure import PhononBandStructureSymmLine

from sumo.electronic_structure.dos import get_pdos

_bohr_to_angstrom = 0.5291772
_ry_to_ev = 13.605693009

# From 2002 CODATA values
to_angstrom = {"ang": 1, "bohr": 0.5291772108, "m": 1e10, "cm": 1e8, "nm": 10}

Tag = collections.namedtuple("Tag", "value comment")
Block = collections.namedtuple("Block", "values comments")

unsupported_dosplot_args = {"elements", "lm_orbitals", "atoms"}


class CastepCell:
    """Structure information and more: CASTEP seedname.cell file

    Usually this will be instantiated with the
    :obj:`sumo.io.castep.CastepCell.from_file()` method

    Keys should always be lower-case; values will retain their case.

    Tags are stored as a dict, where each key is a tag and each value is a
    Tag NamedTuple with the attributes 'value' and 'comment'.

    Blocks are also stored as a dict, where each key is the block name
    (e.g. 'bs_kpoint_list') and the value is a Block NamedTuple with attributes
    'values' (a list of lists) and 'comments' (a commensurate list of str).

    """

    def __init__(self, blocks=None, tags=None):
        self.blocks = blocks if blocks else {}
        self.tags = tags if tags else {}

    @property
    def structure(self):
        # Get lattice vectors
        if "lattice_cart" in self.blocks:
            lattice_cart = self.blocks["lattice_cart"].values
            if lattice_cart[0][0].lower() in ("ang", "nm", "cm", "m", "bohr", "a0"):
                unit = lattice_cart[0][0].lower()
                vectors = lattice_cart[1:]
            else:
                unit = "ang"
                vectors = lattice_cart
            vectors = np.array([list(map(float, row)) for row in vectors])
            if vectors.shape != (3, 3):
                raise ValueError("lattice_cart should contain a 3x3 matrix")

            vectors *= to_angstrom[unit]
            lattice = Lattice(vectors)
        elif "lattice_abc" in self.blocks:
            lattice_abc = self.blocks["lattice_abc"].values
            if lattice_abc[0][0].lower() in ("ang", "nm", "cm", "m", "bohr", "a0"):
                unit = lattice_abc[0][0].lower()
                lengths_and_angles = lattice_abc[1:]
            else:
                unit = "ang"
                lengths_and_angles = lattice_abc[1:]
            if len(lengths_and_angles) != 2:
                raise ValueError("lattice_abc should have two rows")
            lengths_and_angles = [list(map(float, row)) for row in lengths_and_angles]
            lengths_and_angles[0] = [
                x * to_angstrom[unit] for x in lengths_and_angles[0]
            ]
            lattice = Lattice.from_lengths_and_angles(*lengths_and_angles)
        else:
            raise ValueError("Couldn't find a lattice in cell file")

        if "positions_frac" in self.blocks:
            elements_coords = [
                (row[0], list(map(float, row[1:4])))
                for row in self.blocks["positions_frac"].values
            ]
            elements, coords = zip(*elements_coords)
            return Structure(lattice, elements, coords, coords_are_cartesian=False)
        elif "positions_abs" in self.blocks:
            positions_abs = self.blocks["positions_abs"].values
            if positions_abs[0][0].lower() in ("ang", "nm", "cm", "m", "bohr", "a0"):
                unit = positions_abs[0][0].lower()
                positions_abs = positions_abs[1:]
            else:
                unit = "ang"
            elements_coords = [
                (row[0], list(map(float, row[1:4]))) for row in positions_abs
            ]
            elements, coords = zip(*elements_coords)
            return Structure(lattice, elements, coords, coords_are_cartesian=True)
        else:
            raise ValueError("Couldn't find any atom positions in cell file")

    def to_file(self, filename):
        with open(filename, "wt") as f:
            for tag, content in self.tags.items():
                if content.comment in (None, ""):
                    f.write(f"{tag: <24}: {' '.join(content.value)}\n")
                else:
                    f.write(
                        "{: <24}: {: <16} ! {}\n".format(
                            tag, " ".join(content.value), content.comment
                        )
                    )
            for block, content in self.blocks.items():
                f.write(f"\n%block {block}\n")
                if content.comments is None:
                    comments = [""] * len(content.values)
                else:
                    comments = content.comments

                for row, comment in zip(content.values, comments):
                    line = " ".join(map(str, row))
                    if comment != "":
                        line = f"{line: <30} ! {comment}"
                    line += "\n"
                    f.write(line)

                f.write(f"%endblock {block}\n")

    @classmethod
    def from_file(cls, filename):

        with zopen(filename, "rt") as f:
            lines = [line.strip() for line in f]

        # Remove lines which are entirely commented or empty
        def _is_not_empty_line(line):
            if len(line) == 0:
                return False
            elif line[0] in "#;!":
                return False
            elif len(line) > 6 and line[:7] == "COMMENT":
                return False
            else:
                return True

        lines = list(filter(_is_not_empty_line, lines))

        tags, blocks = {}, {}
        current_block_values, current_block_comments = [], []
        in_block = False
        current_block_label = ""

        for line in lines:
            if len(line.split()) == 0:
                continue
            elif line[:6].lower() == "%block":
                if in_block:
                    raise OSError(
                        "Cell file contains nested blocks. "
                        "This possibility was not anticipated."
                    )
                else:
                    current_block_label = line.split()[1].lower()
                    in_block = True
                    continue
            elif line[:9].lower() == "%endblock":
                if line.split()[1].lower() != current_block_label:
                    raise OSError(
                        "Endblock {} does not match current block "
                        "{}: cannot interpret cell file.".format(
                            line.split()[1], current_block_label
                        )
                    )
                if in_block:
                    blocks[current_block_label] = Block(
                        current_block_values, current_block_comments
                    )

                    current_block_values, current_block_comments = [], []
                    in_block = False
                else:
                    raise OSError(
                        f"Cannot cope with line {line}: not currently in a block."
                    )
            elif in_block:
                _, data, comment = _data_comment_from_line(line, in_block=True)
                current_block_values.append(data)
                current_block_comments.append(comment)

            else:
                tag, data, comment = _data_comment_from_line(line)
                tags[tag.lower()] = Tag(data, comment)

        return cls(tags=tags, blocks=blocks)

    @classmethod
    def from_structure(cls, structure):
        blocks = {
            "lattice_cart": Block(
                [["ang"]] + [list(map(str, row)) for row in structure.lattice.matrix],
                ["", "", "", ""],
            ),
            "positions_frac": Block(
                [
                    [site.species_string, *map(str, site.frac_coords)]
                    for site in structure.sites
                ],
                ["" for site in structure.sites],
            ),
        }
        return cls(blocks=blocks)


def read_dos(
    bands_file,
    pdos_file=None,
    cell_file=None,
    bin_width=0.01,
    gaussian=None,
    padding=None,
    emin=None,
    emax=None,
    efermi_to_vbm=True,
    lm_orbitals=None,
    elements=None,
    atoms=None,
    total_only=False,
):
    """Convert DOS data from CASTEP .bands file to Pymatgen/Sumo format

    The data is binned into a regular series using np.histogram

    Args:
        bands_file (:obj:`str`): Path to CASTEP prefix.bands output file. The
            k-point positions, weights and eigenvalues are read from this file.
        bin_width (:obj:`float`, optional): Spacing for DOS energy axis
        gaussian (:obj:`float` or None, optional): Width of Gaussian broadening
            function
        padding (:obj:`float`, optional): Energy range above and below occupied
            region. (This is not used if xmin and xmax are set.)
        emin (:obj:`float`, optional): Minimum energy value for output DOS)
        emax (:obj:`float`, optional): Maximum energy value for output DOS
        efermi_to_vbm (:obj:`bool`, optional):
            If a bandgap is detected, modify the stored Fermi energy
            so that it lies at the VBM.
        elements (:obj:`dict`, optional): The elements and orbitals to extract
            from the projected density of states. Should be provided as a
            :obj:`dict` with the keys as the element names and corresponding
            values as a :obj:`tuple` of orbitals. For example, the following
            would extract the Bi s, px, py and d orbitals::

                {'Bi': ('s', 'px', 'py', 'd')}

            If an element is included with an empty :obj:`tuple`, all orbitals
            for that species will be extracted. If ``elements`` is not set or
            set to ``None``, all elements for all species will be extracted.
        lm_orbitals (:obj:`dict`, optional): The orbitals to decompose into
            their lm contributions (e.g. p -> px, py, pz). Should be provided
            as a :obj:`dict`, with the elements names as keys and a
            :obj:`tuple` of orbitals as the corresponding values. For example,
            the following would be used to decompose the oxygen p and d
            orbitals::

                {'O': ('p', 'd')}

        atoms (:obj:`dict`, optional): Which atomic sites to use when
            calculating the projected density of states. Should be provided as
            a :obj:`dict`, with the element names as keys and a :obj:`tuple` of
            :obj:`int` specifying the atomic indices as the corresponding
            values. The elemental projected density of states will be summed
            only over the atom indices specified. If an element is included
            with an empty :obj:`tuple`, then all sites for that element will
            be included. The indices are 0 based for each element specified in
            the POSCAR. For example, the following will calculate the density
            of states for the first 4 Sn atoms and all O atoms in the
            structure::

                {'Sn': (1, 2, 3, 4), 'O': (, )}

            If ``atoms`` is not set or set to ``None`` then all atomic sites
            for all elements will be considered.

    Returns:
        (:obj:`pymatgen.electronic_structure.dos.Dos`, dict)

        where the dict is either empty or contains a PDOS arranged::

          {species: {orbital: Dos}}
    """

    header = _read_bands_header_verbose(bands_file)

    logging.info("Reading band eigenvalues...")
    _, weights, eigenvalues = read_bands_eigenvalues(bands_file, header)

    calc_efermi = header["e_fermi"][0] * _ry_to_ev * 2
    if efermi_to_vbm and not _is_metal(eigenvalues, calc_efermi):
        logging.info("Setting energy zero to VBM")
        efermi = _get_vbm(eigenvalues, calc_efermi)
    else:
        logging.info("Setting energy zero to Fermi energy")
        efermi = calc_efermi

    emin_data = min(eigenvalues[Spin.up].flatten())
    emax_data = max(eigenvalues[Spin.up].flatten())
    if Spin.down in eigenvalues:
        emin_data = min(emin_data, min(eigenvalues[Spin.down].flatten()))
        emax_data = max(emax_data, max(eigenvalues[Spin.down].flatten()))

    if padding is None and gaussian:
        padding = gaussian * 3
    elif padding is None:
        padding = 0.5

    if emin is None:
        emin = emin_data - padding
    if emax is None:
        emax = emax_data + padding

    # Shift sampling window to account for zeroing at VBM/EFermi
    emin += efermi
    emax += efermi

    bins = np.arange(emin, emax + bin_width, bin_width)
    energies = (bins[1:] + bins[:-1]) / 2

    # Add rows to weights for each band so they are aligned with eigenval data
    weights = weights * np.ones([eigenvalues[Spin.up].shape[0], 1])

    dos_data = {
        spin: np.histogram(eigenvalue_set, bins=bins, weights=weights)[0]
        for spin, eigenvalue_set in eigenvalues.items()
    }

    dos = Dos(efermi, energies, dos_data)

    if pdos_file is not None and not total_only:
        if cell_file is None:
            raise OSError(
                f"Cell file {cell_file} not found: this must be provided for PDOS."
            )
        pdos_raw = compute_pdos(pdos_file, eigenvalues, weights, bins)
        # Also we, need to read the structure, but have it sorted with increasing
        # atomic numbers
        structure = CastepCell.from_file(cell_file).structure.get_sorted_structure(
            key=lambda x: x.species.elements[0].Z
        )
        pdoss = {}
        for isite, site in enumerate(structure.sites):
            pdoss[site] = pdos_raw[isite]
        # Get the pdos dictionary for potting
        pdos = get_pdos(
            CompleteDos(structure, dos, pdoss),
            lm_orbitals=lm_orbitals,
            elements=elements,
            atoms=atoms,
        )
        # Smear the PDOS
        for orbs in pdos.values():
            for dtmp in orbs.values():
                if gaussian:
                    dtmp.densities = dtmp.get_smeared_densities(gaussian)
    else:
        pdos = {}

    if gaussian:
        dos.densities = dos.get_smeared_densities(gaussian)

    return dos, pdos


def _is_metal(eigenvalues, efermi, tol=1e-5):
    # Detect if material is a metal by checking if bands cross efermi
    from itertools import chain

    for band in chain(*eigenvalues.values()):
        if np.any(band < (efermi - tol)) and np.any(band > (efermi + tol)):
            logging.info("Electronic structure appears to be a metal")
            return True

    logging.info("Electronic structure appears to have a bandgap")
    return False


def _get_vbm(eigenvalues, efermi):
    from itertools import chain

    occupied_states_by_band = (
        band[band < efermi] for band in chain(*eigenvalues.values())
    )
    return max(chain(*occupied_states_by_band))


def band_structure(bands_file, cell_file=None):
    """Convert band structure data from CASTEP to Pymatgen/Sumo format

    Args:
        bands_file (:obj:`str`): Path to CASTEP prefix.bands output file. The
            lattice parameters, k-point positions and eigenvalues are read from
            this file.
        cell_file (:obj:`str`, optional): Path to CASTEP prefix.cell input
            file. If found, the positions of high-symmetry points are read from
            the ``[bs|spectral]_kpoint(s)_[path|list]`` block in this file.

    Returns:
        :obj:`pymatgen.electronic_structure.bandstructure.BandStructureSymmLine`

    """

    header = _read_bands_header_verbose(bands_file)

    lattice = Lattice(header["lattice_vectors"])
    lattice = Lattice(lattice.matrix * _bohr_to_angstrom)

    logging.info("Reading band structure eigenvalues...")
    kpoints, _, eigenvalues = read_bands_eigenvalues(bands_file, header)

    if cell_file is None:
        labels = {}
    else:
        labels = labels_from_cell(cell_file)

    return BandStructureSymmLine(
        kpoints,
        eigenvalues,
        lattice.reciprocal_lattice_crystallographic,
        header["e_fermi"][0] * _ry_to_ev * 2,
        labels,
        coords_are_cartesian=False,
    )


def labels_from_cell(cell_file, phonon=False):
    """Read special k-point positions/labels from CASTEP cell file

    Args:
        cell_file (:obj:`str`): Path to CASTEP prefix.cell input file
            high-symmetry points will be read from a block beginning with

                %block bs_kpoint_path

            and ending with

                %endblock bs_kpoint path

            where each special-point line is formatted

                k1 k2 k3  ! label

            where k1, k2, k3 are given in fractional coordinates of the
            reciprocal lattice.
        phonon (:obj:`bool`): Look for phonon interpolation  blocks instead of
            electronic band structure (e.g. PHONON_FINE_KPOINT_PATH)

    returns:
        {:obj:`str`: :obj:`tuple`, ...}

    """

    labels = {}

    if phonon:
        blockstart = re.compile(r"^%block\s+phonon(_fine)?_kpoint(s)?_(path|list)")
        blockend = re.compile(r"^%endblock\s+phonon(_fine)?_kpoint(s)?_(path|list)")
    else:
        blockstart = re.compile(r"^%block\s+(bs|spectral)_kpoint(s)?_(path|list)")
        blockend = re.compile(r"^%endblock\s+(bs|spectral)_kpoint(s)?_(path|list)")

    with zopen(cell_file, "r") as f:
        line = ""
        while blockstart.match(line.lower()) is None:
            line = f.readline()
            if line == "":
                logging.error("Could not find start of k-point block in cell file.")
                sys.exit()

        line = f.readline()  # Skip past block start line
        while blockend.match(line.lower()) is None:
            kpt = tuple(map(float, line.split()[:3]))
            if len(line.split()) > 3:
                label = line.split()[-1]
                if label.lower() in ("g", "gamma"):
                    label = r"\Gamma"
                labels[label] = kpt
            line = f.readline()
            if line == "":
                logging.error("Could not find end of k-point block in cell file.")
                sys.exit()

    return labels


def _read_bands_header_verbose(bands_file):
    logging.info("Reading CASTEP .bands file header...")
    header = read_bands_header(bands_file)

    if header["n_spins"] == 1:
        logging.info(
            "nbands: {}, efermi / Ry: {}, spin channels: 1".format(
                header["n_bands"][0], header["e_fermi"][0]
            )
        )
    elif header["n_spins"] == 2:
        logging.info(
            "nbands: {}, efermi / Ry: {}, spin channels: 2".format(
                header["n_bands"], header["e_fermi"]
            )
        )
    else:
        raise ValueError("Should only be 1 or 2 spin channels!")

    if len(header["e_fermi"]) == 2:
        if abs(header["e_fermi"][1] - header["e_fermi"][1]) > 1e-8:
            raise NotImplementedError(
                "Different Fermi energy in each channel."
                " I have no idea how to handle that."
            )
    return header


def read_bands_header(bands_file):
    """Read CASTEP bands file header, get lattice vectors and metadata

    Args:
        bands_file (:obj:`str`): Path to CASTEP prefix.bands output file

    Returns:
        :obj:`dict`
            CASTEP bands metadata of the form::

            {'lattice': [[ax, ay, az], [bx, by, bz], [cx, cy, cz]],
             'n_kpoints': n_kpoints,
             'n_spins': n_spins,
             'n_electrons': [n_spin1, n_spin2],
             'n_bands': [n_bands_spin1, n_bands_spin2],
             'e_fermi': [e_fermi_spin1, e_fermi_spin2]}

    """

    header = {}
    with zopen(bands_file, "r") as f:
        # Read in header information a line at a time

        header["n_kpoints"] = int(f.readline().split()[-1])
        header["n_spins"] = int(f.readline().split()[-1])
        header["n_electrons"] = [float(x) for x in f.readline().split()[3:]]
        header["n_bands"] = [int(x) for x in f.readline().split()[3:]]

        fermi_line = f.readline().split()
        if fermi_line[3] != "atomic":
            raise NotImplementedError("Band data not in atomic units")
        header["e_fermi"] = [float(x) for x in fermi_line[5:]]

        vectors_header = f.readline().strip()
        if vectors_header != "Unit cell vectors":
            raise AssertionError(
                "CASTEP bands file not formatted as expected", vectors_header
            )

        lattice_vectors = [[float(x) for x in f.readline().split()] for _ in range(3)]
        for row in lattice_vectors:
            if len(row) != 3:
                raise AssertionError("Unit cell vectors not read correctly")
        header["lattice_vectors"] = lattice_vectors

    return header


def read_bands_eigenvalues(bands_file, header):
    """Read eigenvalue data from a CASTEP bands file

    Args:
        bands_file (:obj:`str`): Path to CASTEP prefix.bands output file
        header (:obj:`dict`): Band metadata, as obtained by reading the bands
            file with :obj:`read_bands_header`.

    Returns:
        (:obj:`list`, :obj:`list`, :obj:`np.ndarray`)
            kpoints list, weights and eigenvalue array ordered by
            [band][kpt]. Each k-point is expressed as numpy array of length
            3. Eigenvalues are converted from Ry to eV.

    """

    # Implementation note: pymatgen band structure expects an eigenvalue array
    # ordered [band][kpt] but the data will be read in one kpt at a time as
    # this is the format of the bnd data file.
    #
    # We will build a nested list [[bnd1_kpt1, bnd2_kpt1, ...],
    #                              [bnd1_kpt2, bnd2_kpt2, ...], ...]
    # then convert to a numpy array (i.e. with each row for a kpoint)
    # and then transpose the array to obtain desired formats
    #
    # CASTEP can write the k-points out-of-order, so we need to keep track of
    # the k-point indices and re-order the lists before array conversion.

    kpoints = []
    if header["n_spins"] == 1:
        eigenvals = {Spin.up: []}
    else:
        eigenvals = {Spin.up: [], Spin.down: []}

    with zopen(bands_file, "r") as f:
        for _ in range(9):
            f.readline()  # Skip header

        for _ in range(header["n_kpoints"]):
            # The first "k-point" column is actually the sorting index
            # and the last is the weighting
            kpoints.append(np.array([float(x) for x in f.readline().split()[1:]]))
            f.readline()  # Skip past "Spin component 1"
            spin1_bands = [kpoints[-1][0]] + [  # Sorting key goes in col 0
                float(f.readline()) for _i in range(header["n_bands"][0])
            ]
            eigenvals[Spin.up].append(spin1_bands)

            if header["n_spins"] == 2:
                f.readline()  # Skip past "Spin component 2"
                spin2_bands = [kpoints[-1][0]] + [  # Sorting key goes in col 0
                    float(f.readline()) for _i in range(header["n_bands"][0])
                ]
                eigenvals[Spin.down].append(spin2_bands)

    # Sort kpoints and trim off sort keys, weights
    kpoint_list = sorted(kpoints, key=(lambda k: k[0]))
    kpoints = [k[1:4] for k in kpoint_list]
    weights = [k[-1] for k in kpoint_list]

    # Sort eigenvalues and trim off sort keys
    for key, data in eigenvals.items():
        data = np.array(data)
        eigenvals[key] = data[data[:, 0].argsort(), 1:]
    # Transpose matrix to arrange by band and convert to eV from Ha
    eigenvals = {key: data.T * _ry_to_ev * 2 for key, data in eigenvals.items()}

    return kpoints, weights, eigenvals


def write_kpoint_files(
    filename,
    kpoints,
    labels,
    make_folders=False,
    kpts_per_split=None,
    phonon=False,
    directory=None,
):
    r"""Write the k-points data to files.

    Folders are named as 'split-01', 'split-02', etc ...
    .cell files are named band_split_01.cell etc ...

    Cartesian coordinates are not supported by CASTEP for band paths.

    Unlike VASP, CASTEP band structure calculations can deal with two
    separate k-point meshes: a mesh for the SCF convergence and a series of
    points (or path segments) for the band structure calculation. It is assumed
    that the existing .cell file contains a specification of the SCF k-point
    mesh, and Sumo will append the band path information.

    Args:
        filename (:obj:`str`): Path to CASTEP structure (.cell) file.
        kpoints (:obj:`numpy.ndarray`): The k-point coordinates along the
            high-symmetry path. For example::

                [[0, 0, 0], [0.25, 0, 0], [0.5, 0, 0], [0.5, 0, 0.25],
                [0.5, 0, 0.5]]

        labels (:obj:`list`) The high symmetry labels for each k-point (will be
            an empty :obj:`str` if the k-point has no label). For example::

                ['\Gamma', '', 'X', '', 'Y']

        make_folders (:obj:`bool`, optional): Generate folders and copy in
            param file from the current directory, setting the TASK and
            SPECTRAL_TASK tags as appropriate.  Cell file *filename* will be
            copied with the new band path block.  If a matching .check file
            exists this will also be copied, and the REUSE tag will be set.

        kpts_per_split (:obj:`int`, optional): If set, the k-points are split
            into separate k-point files (or folders) each containing the number
            of k-points specified. This is useful for hybrid band structure
            calculations where it is often intractable to calculate all
            k-points in the same calculation.

        phonon (:obj:`bool`, optional): Write the Brillouin-zone path as an
            interpolated phonon q-point path (PHONON_FINE_KPOINTS_PATH) instead
            of electronic band-structure k-points.

        directory (:obj:`str`, optional): The output file directory.
    """

    if kpts_per_split:
        kpt_splits = [
            kpoints[i : i + kpts_per_split]
            for i in range(0, len(kpoints), kpts_per_split)
        ]
        label_splits = [
            labels[i : i + kpts_per_split]
            for i in range(0, len(labels), kpts_per_split)
        ]
    else:
        kpt_splits = [kpoints]
        label_splits = [labels]

    kpt_cell_files = []

    if phonon:
        tags = {"task": "Phonon"}
        kpoint_tag = "phonon_fine_kpoint_list"
        clash_tags = {
            "phonon_fine_kpoints_list",
            "phonon_fine_kpoint_path",
            "phonon_fine_kpoints_path",
        }
    else:
        tags = {"task": "Spectral", "spectral_task": "BandStructure"}
        kpoint_tag = "spectral_kpoint_list"
        clash_tags = {
            "{}_{}_{}".format(*tag_parts)
            for tag_parts in product(
                ("bs", "spectral"),
                ("kpoint", "kpoints"),
                ("path", "list", "mp_grid", "mp_spacing", "mp_offset"),
            )
        }

    for kpt_split, label_split in zip(kpt_splits, label_splits):
        cellfile = CastepCell.from_file(filename)

        # Remove any spectral blocks which may have already been loaded
        for key in clash_tags:
            try:
                del cellfile.blocks[key]
            except KeyError:
                pass

        cellfile.blocks[kpoint_tag] = Block(kpt_split, label_split)
        kpt_cell_files.append(cellfile)

    pad = int(math.floor(math.log10(len(kpt_cell_files)))) + 2
    if make_folders:
        # Derive param file name by changing extension from cell file
        seedname = ".".join(filename.split(".")[:-1])
        param_file = seedname + ".param"
        check_file = seedname + ".check"

        for i, cell_file in enumerate(kpt_cell_files):
            split_folder = f"split-{str(i + 1).zfill(pad)}"
            if directory:
                split_folder = os.path.join(directory, split_folder)
            os.mkdir(split_folder)

            if os.path.isfile(check_file):
                shutil.copy(check_file, split_folder)
                tags.update({"reuse": os.path.basename(check_file)})

            copy_param(param_file, split_folder, tags=tags)

            cell_file.to_file(os.path.join(split_folder, os.path.basename(filename)))

    else:
        for i, cell_file in enumerate(kpt_cell_files):
            if len(kpt_cell_files) > 1:
                cell_filename = f"band_split_{i + 1:0d}.cell"
            else:
                cell_filename = "band.cell"
            if directory:
                cell_filename = os.path.join(directory, cell_filename)
            cell_file.to_file(cell_filename)


def copy_param(filename, folder, tags=None):
    """Copy CASTEP .param file, modifying Task if needed

    Args:
        filename (:obj:`str`): Path to CASTEP .param file
        folder (:obj:`str`): Directory for output .param file
        tags (:obj:`dict` or :obj:`None`):
            Tags to replace or insert (e.g. {'task': 'Phonon'})

    """

    output_filename = os.path.join(folder, os.path.basename(filename))
    if os.path.isfile(filename) and tags is None:
        shutil.copyfile(filename, output_filename)

    elif os.path.isfile(filename):
        # Copy tags dict and remove case-sensitive keys
        tags = {key.lower(): value for key, value in tags.items()}

        with open(filename) as infile:
            with open(output_filename, "w") as outfile:
                for line in infile:
                    tag = line.split()[0].lower() if line.split() else None
                    if tag in tags:
                        outfile.write(f"{tag} : {tags.pop(tag)}\n")
                    else:
                        outfile.write(line)
                # Write remaining tags
                for tag, value in tags.items():
                    outfile.write(f"{tag} : {value}\n")
    else:
        logging.warning(f"Cannot find param file {filename}, skipping")


def _data_comment_from_line(line, in_block=False):
    """Parse a line from a CASTEP input file

    Args:
        line (str): line from CASTEP .cell or .param file
        in_block (bool): Include all tokens before comment in
            data and leave tag as None.

    Returns:
        (tag, data, comment) where tag is a str, data is a list
        of str and comment is a str

    """
    comment_split_line = re.split("[#;!]", line)

    line_content = comment_split_line[0]
    if len(comment_split_line) == 1:
        comment = ""
    else:
        comment = line[len(line_content) :]

    if in_block:
        tag, data = None, line.split()
    else:
        tag, *data = line_content.split()
        if len(data) == 0:
            data = [""]

        # Clean up delimiter: 'TAG DAT' 'TAG= DAT' 'TAG : DAT'
        # are all possible so check both symbols in both places
        data = [item for item in data if item not in ":="]
        if tag[-1] in ":=":
            tag = tag[:-1]

    return tag, data, comment


class CastepPhonon:
    """Data from CASTEP phonon calculation: seedname.phonon file

    Usually this will be instantiated with the
    :obj:`sumo.io.castep.CastepPhonon.from_file()` method

    """

    def __init__(self, header, qpts, frequencies, weights=None, eigenvectors=None):
        """
        Args:
            header (:obj:`dict`):
                Dict containing key metadata from header file. Positions are in
                fractional coordinates.

                  {nions: NIONS, nbranches: NBRANCHES, nqpts: NQPTS,
                   frequency_unit: 'cm-1',
                   cell: [[ax, ay, az], [bx, by, bz], [cx, cy, cz]],
                   positions: [[a1, b1, c1], [a2, b2, c2], ...],
                   symbols: [el1, el2, ...],
                   masses: [m1, m2, ...]}

            qpts (:obj:`numpy.ndarray`):
                Nx3 array of q-points in fractional coordinates
            frequencies (:obj:`numpy.ndarray`):
                2-D array of frequencies arranged [mode, qpt]. (Frequencies are
                in units of header['frequency_unit']).
            weights (:obj:`numpy.ndarray`, optional):
                1-D array of q-point weights (not used in band structure plot).
            eigenvectors (:obj:`numpy.ndarray`, optional):
                4-D array of phonon eigenvectors arranged
                [mode, qpt, atom, xyz] (not used in band structure plot).

        """
        self.header = header
        self.qpts = qpts
        self.weights = weights
        self.frequencies = frequencies
        self.eigenvectors = eigenvectors
        self.labels = {}

    @classmethod
    def from_file(cls, filename):
        """Create a CastepPhonon object by reading CASTEP .phonon file

        Args:
            filename (:obj:`str`): Input .phonon file

        Returns:
            pymatgen.phonon.bandstructure.PhononBandStructure
        """
        header = read_phonon_header(filename)
        qpts, weights, frequencies, eigenvectors = read_phonon_bands(filename, header)

        return cls(
            header, qpts, frequencies, weights=weights, eigenvectors=eigenvectors
        )

    def set_labels_from_file(self, filename):
        """Set dictionary of special-point labels from .cell file comments

        Args:
            filename (:obj:`str`):
                Input .cell file. Special point symbols should be included as
                comments on the lines where those qpts are defined; *kgen* will
                do this automatically when writing high-symmetry paths.

        """
        self.labels = labels_from_cell(filename, phonon=True)

    def get_band_structure(self):
        lattice = Lattice(self.header["cell"])
        structure = Structure(
            lattice, species=self.header["symbols"], coords=self.header["positions"]
        )

        # I think this is right? Need to test somehow...
        mass_weights = 1 / np.sqrt(self.header["masses"])
        displacements = (
            self.eigenvectors * mass_weights[np.newaxis, np.newaxis, :, np.newaxis]
        )

        return PhononBandStructureSymmLine(
            self.qpts,
            self.frequencies,
            lattice.reciprocal_lattice,
            structure=structure,
            eigendisplacements=displacements,
            labels_dict=self.labels,
        )


def read_phonon_header(filename):
    """Read header information from CASTEP .phonon Args

    "Number of wavevectors" reported in the heading can be misleading as LO-TO
    splitting adds extra Gamma-point entries. To get the array sizes straight
    we also search the file for a "true N-qpts".

    File:
        filename (:obj:`str`): Input file

    Returns:
        :obj:`dict`:

        Dict containing key metadata from header file. Positions are in
        fractional coordinates.

          {n_ions: NIONS, n_branches: NBRANCHES, n_qpts: NQPTS,
           n_entries: NQPTS(INCLUDING LOTO DUPLICATES),
           frequency_unit: 'cm-1',
           cell: [[ax, ay, az], [bx, by, bz], [cx, cy, cz]],
           positions: [[a1, b1, c1], [a2, b2, c2], ...],
           symbols: [el1, el2, ...],
           masses: [m1, m2, ...]}
    """

    int_keys = {"ions": "n_ions", "branches": "n_branches", "wavevectors": "n_qpts"}

    header = {}
    in_header = False
    with zopen(filename, "rt") as f:
        for line in f:
            line = line.strip()
            if line[:12] == "BEGIN header":
                in_header = True
            elif line[:10] == "END header":
                break
            elif not in_header:
                pass
            else:
                split_line = line.split()
                # Get int data e.g. "Number of ions      1"
                if split_line[0] == "Number":
                    header[int_keys[split_line[2]]] = int(split_line[3])

                # Lattice vector block: next() will skip lines from for-loop
                elif split_line[:3] == ["Unit", "cell", "vectors"]:
                    logging.info("Reading unit cell...")
                    cell = [next(f).split(), next(f).split(), next(f).split()]
                    header["cell"] = np.array(cell, dtype=float).tolist()

                # Positions block: pre-allocate arrays and fill line-by-line
                elif split_line[:2] == ["Fractional", "Co-ordinates"]:
                    if not header.get("n_ions"):
                        raise OSError(
                            "Could not find number of ions; phonon "
                            "file not formatted correctly."
                        )

                    logging.info("Reading structure...")

                    header["positions"] = []
                    header["symbols"] = []
                    header["masses"] = []

                    for _ in range(header["n_ions"]):
                        _, *position, el, m = next(f).split()
                        header["positions"].append(list(map(float, position)))
                        header["symbols"].append(el)
                        header["masses"].append(float(m))
                else:
                    logging.debug(f'Skipping unused header line "{line}"')

        logging.info("Finished reading header, checking number of entries ...")
        is_qpt = re.compile(r"^\s+q-pt=\s+")
        header["n_entries"] = sum(bool(is_qpt.match(line)) for line in f)
        if header["n_entries"] > header["n_qpts"]:
            logging.info(
                "Found more entries than reported q-points: "
                "assume LO-TO splitting was used"
            )

    return header


def read_phonon_bands(filename, header):
    """Using header information, read band data from CASTEP .phonon file

    Args:
        filename (:obj:`str`): Input file
        header (:obj:`dict`): Header data including

          {'n_ions': NIONS, 'n_branches': NBRANCHES, 'n_entries': ENTRIES, ...}

          Note that ENTRIES may be larger than the reported "Number of
          wavevectors" as Gamma points are sometimes duplicated to show LO-TO
          splitting.

    Returns:
        :obj:`tuple` (qpts, weights, frequencies, eigenvectors)

        where *qpts* is a 3xN :obj:`numpy.ndarray` of q-points in fractional
        coordinates, *weights* is a 1-D :obj:`numpy.ndarray` of q-point
        weights, *frequencies* is a 2-D :obj:`numpy.ndarray` of frequencies
        arranged [mode, qpt] and *eigenvectors* is a 4-D :obj:`numpy.ndarray`
        of complex phonon eigenvectors arranged [mode, qpt, atom, xyz].

    """
    logging.info("Reading phonon qpts, frequencies, weights, eigenvectors...")

    # Pre-allocate arrays
    qpts = np.zeros((header["n_entries"], 3))
    weights = np.zeros(header["n_entries"])
    frequencies = np.zeros((header["n_branches"], header["n_entries"]))
    eigenvectors = np.zeros(
        (header["n_branches"], header["n_entries"], header["n_ions"], 3),
        dtype=np.complex64,
    )

    # Scan past header
    with zopen(filename, "rt") as f:
        for line in f:
            if line.strip() == "END header":
                break
        else:
            raise OSError(
                'Did not find "END header" line in phonon file "{}".'
                " File is not formatted correctly.".format(filename)
            )

        # Read data blocks
        for i_qpt in range(header["n_entries"]):
            # Typical q-pt line format:
            # q-pt=   1   0.0000  0.0000  0.000   0.003012
            #
            #  -    INDEX ---------QPT---------    WEIGHT
            #  0      1      2      3       4         5

            qpt_line = f.readline().split()
            if not qpt_line[0] == "q-pt=":
                raise AssertionError()

            # We can't check if the indices are correct because CASTEP gives a
            # duplicate index to "Gamma-point" with delta for LO-TO splitting,
            # so skip over the qpt number and go to data

            qpts[i_qpt, :] = list(map(float, qpt_line[2:5]))
            weights[i_qpt] = float(qpt_line[5])

            # Next come frequencies, typical line   "   1    3.25123"
            for i_mode in range(header["n_branches"]):
                frequencies[i_mode, i_qpt] = float(f.readline().split()[1])

            # Next come eigenvectors, with some nice tabulation comments, e.g.
            #
            #                          Phonon Eigenvectors
            #   Mode Ion         X                Y               Z
            #      1   1  0.54308  0.0000  -0.7995  0.0000  0.2564  0.0000
            #      2   1 -0.60969  0.0000  -0.1654  0.0000  0.7751  0.0000
            if not f.readline().strip() == "Phonon Eigenvectors":
                raise AssertionError()
            if "X" not in f.readline().split():
                raise AssertionError()
            for i_mode in range(header["n_branches"]):
                for i_ion in range(header["n_ions"]):
                    eigenvector = np.fromstring(f.readline(), sep=" ")[2:]
                    eigenvector = eigenvector[::2] + 1j * eigenvector[1::2]
                    eigenvectors[i_mode, i_qpt, i_ion, :] = eigenvector

        return qpts, weights, frequencies, eigenvectors



# Copyright (c) Scanlon Materials Theory Group
# Distributed under the terms of the MIT License.

"""
A script to calculate and plot optical spectra from ab initio calculations.
"""

import argparse
import logging
import os
import sys
import warnings
from collections import OrderedDict
from glob import glob

import matplotlib as mpl

mpl.use("Agg")

from pymatgen.io.vasp import Vasprun
from pymatgen.util.string import latexify

from sumo.electronic_structure.optics import (
    broaden_eps,
    calculate_dielectric_properties,
    write_files,
)
from sumo.io import questaal
from sumo.plotting.optics_plotter import SOpticsPlotter

__author__ = "Alex Ganose"
__version__ = "1.0"
__maintainer__ = "Alex Ganose"
__email__ = "alexganose@googlemail.com"
__date__ = "Jan 10, 2018"


def optplot(
    modes=("absorption",),
    filenames=None,
    codes="vasp",
    prefix=None,
    directory=None,
    gaussian=None,
    band_gaps=None,
    labels=None,
    average=True,
    height=6,
    width=6,
    xmin=0,
    xmax=None,
    ymin=0,
    ymax=1e5,
    colours=None,
    style=None,
    no_base_style=None,
    image_format="pdf",
    dpi=400,
    plt=None,
    fonts=None,
    units="eV",
):
    """A script to plot optical absorption spectra from VASP calculations.

    Args:
        modes (:obj:`list` or :obj:`tuple`):
            Ordered list of :obj:`str` determining properties to plot.
            Accepted options are 'absorption' (default), 'eps', 'eps-real',
            'eps-im', 'n', 'n-real', 'n-im', 'loss' (equivalent to n-im).
        filenames (:obj:`str` or :obj:`list`, optional): Path to data file.
            For VASP this is a *vasprun.xml* file (can be gzipped); for
            Questaal the *opt.ext* file from *lmf* or *eps_BSE.out* from
            *bethesalpeter* may be used.
            Alternatively, a list of paths can be
            provided, in which case the absorption spectra for each will be
            plotted concurrently.
        codes (:obj:`str` or :obj:`list`, optional): Original
            calculator. Accepted values are 'vasp' and 'questaal'. Items should
            correspond to filenames.
        prefix (:obj:`str`, optional): Prefix for file names.
        directory (:obj:`str`, optional): The directory in which to save files.
        gaussian (:obj:`float`): Standard deviation for gaussian broadening.
        band_gaps (:obj:`float`, :obj:`str` or :obj:`list`, optional): The band
            gap as a :obj:`float`, in eV, plotted as a dashed line. If plotting
            multiple spectra then a :obj:`list` of band gaps can be provided.
            Band gaps can be provided as a floating-point number or as a path
            to a *vasprun.xml* file. To skip over a line, set its bandgap to
            zero or a negative number to place it outside the visible range.
        labels (:obj:`str` or :obj:`list`): A label to identify the spectra.
            If plotting multiple spectra then a :obj:`list` of labels can
            be provided.
        average (:obj:`bool`, optional): Average the dielectric response across
            all lattice directions. Defaults to ``True``.
        height (:obj:`float`, optional): The height of the plot.
        width (:obj:`float`, optional): The width of the plot.
        xmin (:obj:`float`, optional): The minimum energy on the x-axis.
        xmax (:obj:`float`, optional): The maximum energy on the x-axis.
        ymin (:obj:`float`, optional): The minimum absorption intensity on the
            y-axis.
        ymax (:obj:`float`, optional): The maximum absorption intensity on the
            y-axis.
        colours (:obj:`list`, optional): A :obj:`list` of colours to use in the
            plot. The colours can be specified as a hex code, set of rgb
            values, or any other format supported by matplotlib.
        style (:obj:`list` or :obj:`str`, optional): (List of) matplotlib style
            specifications, to be composed on top of Sumo base style.
        no_base_style (:obj:`bool`, optional): Prevent use of sumo base style.
            This can make alternative styles behave more predictably.
        image_format (:obj:`str`, optional): The image file format. Can be any
            format supported by matplotlib, including: png, jpg, pdf, and svg.
            Defaults to pdf.
        dpi (:obj:`int`, optional): The dots-per-inch (pixel density) for
            the image.
        plt (:obj:`matplotlib.pyplot`, optional): A
            :obj:`matplotlib.pyplot` object to use for plotting.
        fonts (:obj:`list`, optional): Fonts to use in the plot. Can be a
            a single font, specified as a :obj:`str`, or several fonts,
            specified as a :obj:`list` of :obj:`str`.
        units (:obj:`str`, optional): X-axis units for the plot. 'eV' for
           energy in electronvolts or 'nm' for wavelength in nanometers.
           Defaults to 'eV'.

    Returns:
        A matplotlib pyplot object.
    """

    # Don't write files if this is being done to manipulate existing plt
    save_files = False if plt else True

    # BUILD LIST OF FILES AUTOMATICALLY IF NECESSARY

    if codes == "vasp":
        if not filenames:
            if os.path.exists("vasprun.xml"):
                filenames = ["vasprun.xml"]
            elif os.path.exists("vasprun.xml.gz"):
                filenames = ["vasprun.xml.gz"]
            else:
                logging.error("ERROR: No vasprun.xml found!")
                sys.exit()

    elif codes == "questaal":
        if not filenames:
            if len(glob("opt.*")) > 0:
                filenames = glob("opt.*")
                if len(filenames) == 1:
                    logging.info("Found optics file: " + filenames[0])
                else:
                    logging.info("Found optics files: " + ", ".join(filenames))

    if isinstance(filenames, str):
        filenames = [filenames]

    if isinstance(codes, str):
        codes = [codes] * len(filenames)
    elif len(codes) == 1:
        codes = list(codes) * len(filenames)

    # ITERATE OVER FILES READING DIELECTRIC DATA

    dielectrics = []
    auto_labels = []
    auto_band_gaps = []
    for i, (filename, code) in enumerate(zip(filenames, codes)):
        if code == "vasp":
            vr = Vasprun(filename)
            dielectrics.append(vr.dielectric)

            auto_labels.append(
                latexify(vr.final_structure.composition.reduced_formula).replace(
                    "$_", r"$_\mathregular"
                )
            )

            if isinstance(band_gaps, list) and not band_gaps:
                # band_gaps = [], auto band gap requested
                auto_band_gaps.append(vr.get_band_structure().get_band_gap()["energy"])
            else:
                auto_band_gaps.append(None)

        elif code == "questaal":
            if not save_files:
                out_filename = None
            elif len(filenames) == 1:
                out_filename = "dielectric.dat"
            else:
                out_filename = f"dielectric_{i + 1}.dat"

            dielectrics.append(questaal.dielectric_from_file(filename, out_filename))

            auto_band_gaps.append(None)
            auto_labels.append(filename.split(".")[-1])
            if isinstance(band_gaps, list) and not band_gaps:
                logging.info(
                    "Bandgap requested but not supported for Questaal"
                    " file {}: skipping...".format(filename)
                )

        else:
            raise Exception(f'Code selection "{code}" not recognised')

    if not labels and len(filenames) > 1:
        labels = auto_labels

    # PROCESS DIELECTRIC DATA: BROADENING AND DERIVED PROPERTIES

    if gaussian:
        dielectrics = [broaden_eps(d, gaussian) for d in dielectrics]

    # initialize spectrum data ready to append from each dataset
    abs_data = OrderedDict()

    for mode in modes:
        abs_data.update({mode: []})

    # for each calculation, get all required properties and append to data
    for d in dielectrics:
        # TODO: add support for other eigs and full modes
        energies, properties = calculate_dielectric_properties(
            d, set(modes), mode="average" if average else "trace"
        )
        for mode, spectrum in properties.items():
            abs_data[mode].append((energies, spectrum))

    if isinstance(band_gaps, list) and not band_gaps:
        # empty list therefore use bandgaps collected from vasprun files
        band_gaps = auto_band_gaps
    elif isinstance(band_gaps, list):
        # list containing filenames and/or values: mutate the list in-place
        for i, item in enumerate(band_gaps):
            if item is None:
                pass
            elif _floatable(item):
                band_gaps[i] = float(item)
            elif "vasprun" in item:
                band_gaps[i] = (
                    Vasprun(item).get_band_structure().get_band_gap()["energy"]
                )
            else:
                raise ValueError(f"Format not recognised for auto bandgap: {item}.")

    plotter = SOpticsPlotter(abs_data, band_gap=band_gaps, label=labels)
    plt = plotter.get_plot(
        width=width,
        height=height,
        xmin=xmin,
        xmax=xmax,
        ymin=ymin,
        ymax=ymax,
        colours=colours,
        dpi=dpi,
        plt=plt,
        fonts=fonts,
        style=style,
        no_base_style=no_base_style,
        units=units,
    )

    if save_files:
        basename = "absorption"
        if prefix:
            basename = f"{prefix}_{basename}"
        image_filename = f"{basename}.{image_format}"

        if directory:
            image_filename = os.path.join(directory, image_filename)
        plt.savefig(image_filename, format=image_format, dpi=dpi)
        for mode, data in abs_data.items():
            basename = "absorption" if mode == "abs" else mode
            write_files(data, basename=basename, prefix=prefix, directory=directory)
    else:
        return plt


def _floatable(item):
    """Check if an item can be intepreted with float()"""
    try:
        float(item)
        return True
    except ValueError:
        return False


def _get_parser():
    parser = argparse.ArgumentParser(
        description="""
    optplot is a script to produce optical absorption spectra diagrams""",
        epilog=f"""
    Author: {__author__}
    Version: {__version__}
    Last updated: {__date__}""",
    )

    parser.add_argument(
        "mode",
        type=str,
        nargs="*",
        default="absorption",
        metavar="M",
        choices={"absorption", "loss", "eps_real", "eps_imag", "n_real", "n_imag"},
        help="Optical properties to plot. Multiple choices "
        " will be displayed as subplots. Accepted values:"
        ' "absorption" (optical absorption over distance)'
        ', "loss" (energy-loss function -Im(1/eps)), '
        '"eps_real" and "eps_imag" (real and imaginary '
        "parts of the dielectric function), "
        '"n_real" (real part of complex refractive index)'
        '"n_imag" (imaginary part of RI, also known as '
        "the extinction coefficient kappa.)",
    )
    parser.add_argument(
        "-f",
        "--filenames",
        metavar="F",
        help="path to one or more vasprun.xml files",
        default=None,
        nargs="+",
    )
    parser.add_argument(
        "-p", "--prefix", metavar="P", help="prefix for the files generated"
    )
    parser.add_argument(
        "-d", "--directory", metavar="D", help="output directory for files"
    )
    parser.add_argument(
        "-c",
        "--code",
        metavar="C",
        default="vasp",
        nargs="+",
        help=("Original calculator. Accepted values are " '"vasp" and "questaal".'),
    )
    parser.add_argument(
        "-g",
        "--gaussian",
        type=float,
        metavar="G",
        help="standard deviation of gaussian broadening",
    )
    parser.add_argument(
        "-b",
        "--bandgaps",
        nargs="*",
        metavar="E",
        help=(
            "indicate the fundamental band gap (options: "
            "nothing, vasprun.xml file, or float). A "
            "sequence of files and values may be provided, "
            "corresponding to the optical data files. "
            "To skip a line, set a value outside the plot "
            "range (e.g. -1)."
        ),
    )
    parser.add_argument(
        "-l",
        "--labels",
        nargs="+",
        metavar="L",
        help="labels for the absorption specta",
    )
    parser.add_argument(
        "-a",
        "--anisotropic",
        action="store_false",
        help="separate spectra into to x, y, and z directions",
    )
    parser.add_argument(
        "--height", type=float, default=None, help="height of the graph"
    )
    parser.add_argument("--width", type=float, default=None, help="width of the graph")
    parser.add_argument(
        "--xmin", type=float, default=0.0, help="minimum energy on the x-axis"
    )
    parser.add_argument(
        "--xmax", type=float, default=None, help="maximum energy on the x-axis"
    )
    parser.add_argument(
        "--ymin",
        type=str,
        default=["auto"],
        nargs="+",
        help="minimum intensity on the y-axis; may specify "
        "multiple values if plotting more than one axis. "
        'Use "auto" or "_" for automatic value.',
    )
    parser.add_argument(
        "--ymax",
        type=str,
        default=["auto"],
        nargs="+",
        help="maximum intensity on the y-axis; may specify"
        "multiple values if plotting more than one axis. "
        'Use "auto" or "_" for automatic value.',
    )
    parser.add_argument(
        "--style",
        type=str,
        nargs="+",
        default=None,
        help="matplotlib style specifications",
    )
    parser.add_argument(
        "--no-base-style",
        action="store_true",
        dest="no_base_style",
        help="prevent use of sumo base style",
    )
    parser.add_argument(
        "--format",
        type=str,
        default="pdf",
        dest="image_format",
        metavar="FORMAT",
        help="image file format (options: pdf, svg, jpg, png)",
    )
    parser.add_argument(
        "--dpi", type=int, default=400, help="pixel density for image file"
    )
    parser.add_argument("--font", default=None, help="font to use")
    parser.add_argument(
        "--units", type=str, default="eV", help="x-axis units (options: eV or nm)"
    )
    return parser


def main():
    args = _get_parser().parse_args()

    logging.basicConfig(
        filename="sumo-optplot.log",
        level=logging.INFO,
        filemode="w",
        format="%(message)s",
    )
    console = logging.StreamHandler()
    logging.info(" ".join(sys.argv[:]))
    logging.getLogger("").addHandler(console)

    warnings.filterwarnings("ignore", category=UserWarning, module="matplotlib")
    warnings.filterwarnings("ignore", category=UnicodeWarning, module="matplotlib")
    warnings.filterwarnings("ignore", category=UserWarning, module="pymatgen")

    # Wrap mode into list if necessary
    if not isinstance(args.mode, list):
        args.mode = [args.mode]

    # Replace text placeholders with preferred Python representation: None
    ymin = [None if (x.lower() in ("auto", "_")) else float(x) for x in args.ymin]
    ymax = [None if (x.lower() in ("auto", "_")) else float(x) for x in args.ymax]

    # Settings should be list corresponding to n_plots, or value for all
    ymin = ymin[0] if len(ymin) == 1 else ymin
    ymax = ymax[0] if len(ymax) == 1 else ymax

    optplot(
        modes=args.mode,
        filenames=args.filenames,
        codes=args.code,
        prefix=args.prefix,
        directory=args.directory,
        gaussian=args.gaussian,
        band_gaps=args.bandgaps,
        labels=args.labels,
        average=args.anisotropic,
        height=args.height,
        width=args.width,
        xmin=args.xmin,
        xmax=args.xmax,
        ymin=ymin,
        ymax=ymax,
        colours=None,
        image_format=args.image_format,
        dpi=args.dpi,
        style=args.style,
        no_base_style=args.no_base_style,
        fonts=args.font,
        units=args.units,
    )


if __name__ == "__main__":
    main()



# Copyright (c) Scanlon Materials Theory Group
# Distributed under the terms of the MIT License.

"""
Module containing functions to process dielectric and optical absorption data.

TODO:
    * Remove magic values
"""

import os

import numpy as np
from scipy.ndimage.filters import gaussian_filter1d


def broaden_eps(dielectric, sigma):
    """Apply gaussian broadening to the dielectric response.

    Args:
        dielectric (tuple): The high-frequency dielectric data, following
            the same format as
            :attr:`pymatgen.io.vasp.outputs.Vasprun.dielectric`.
            This is a :obj:`tuple` containing the energy, the real part of the
            dielectric tensor, and the imaginary part of the tensor, as a
            :obj:`list` of :obj:`floats`. E.g.::

                (
                    [energies],
                    [[real_xx, real_yy, real_zz, real_xy, real_yz, real_xz]],
                    [[imag_xx, imag_yy, imag_zz, imag_xy, imag_yz, imag_xz]]
                )

        sigma (float): Standard deviation for gaussian broadening.

    Returns:
        :obj:`tuple` of :obj:`list` of :obj:`list` of :obj:`float`: The
        broadened dielectric response. Returned as a tuple containing the
        energy, the real part of the dielectric tensor, and the imaginary
        part of the tensor. E.g.::

            (
                [energies],
                [[real_xx, real_yy, real_zz, real_xy, real_yz, real_xz]],
                [[imag_xx, imag_yy, imag_zz, imag_xy, imag_yz, imag_xz]]
            )
    """
    e = dielectric[0]
    diff = [e[i + 1] - e[i] for i in range(len(e) - 1)]
    diff_avg = sum(diff) / len(diff)
    real = [
        gaussian_filter1d(np.array(dielectric[1])[:, x], sigma / diff_avg)
        for x in range(6)
    ]
    imag = [
        gaussian_filter1d(np.array(dielectric[2])[:, x], sigma / diff_avg)
        for x in range(6)
    ]

    return e, np.array(real).T, np.array(imag).T


def calculate_dielectric_properties(dielectric, properties, mode="average"):
    r"""Calculate optical properties from the dielectric function

    Supported properties:

    *Absorption*

    The unit of alpha is :math:`\mathrm{cm}^{-1}`.

    Refractive index :math:`n` has real and imaginary parts:

    .. math::

        n = [(e^\prime + ie^{\prime\prime} / e_0]^{1/2}
          = n^\prime + in^{\prime\prime}

    Relationship between :math:`a` and imaginary :math:`n^{\prime\prime}`:

    .. math::

        a = 4 \pi n^{\prime\prime} / \lambda

    Where:

    .. math:: \lambda = hc/E

    Args:
        dielectric (tuple): The high-frequency dielectric data, following
            the same format as :obj:`pymatgen.io.vasp.Vasprun.dielectric`.
            This is a :obj:`tuple` containing the energy, the real part of the
            dielectric tensor, and the imaginary part of the tensor, as a
            :obj:`list` of :obj:`floats`. E.g.::

                (
                    [energies],
                    [[real_xx, real_yy, real_zz, real_xy, real_yz, real_xz]],
                    [[imag_xx, imag_yy, imag_zz, imag_xy, imag_yz, imag_xz]]
                )

        properties (set):
            The set of properties to return. Intermediate properties will be
            calculated as needed. Accepted values: 'eps_real', 'eps_imag',
            'absorption', 'loss', 'n_real', 'n_imag'

        mode (str): The format of the results: Options are:

            - "average": Average the dielectric response for all directions.
            - "trace": The trace of the dielectric tensor (i.e., along xx, yy, zz).
            - "full": The full dielectric tensor (i.e., xx, xy, xz, yx, yy, yz, zx, zy,
              zz).
            - "eigs": Calculate the eigenvalues of the dielectric response. The
              eigenvalues are sorted from smallest to largest.

    Returns:
        tuple of ``(energies, {property_name: property_value})``: The optical absorption is given in
        :math:`\mathrm{cm}^{-1}`. If ``mode`` is ``"average"``, the
        property_value will be returned as::

            [property]

        If ``mode`` is ``"trace"``, the data will be returned as::

            [property_xx, property_yy, property_zz]

        If ``mode`` is ``"full"``, the data will be returned as::

            [xx, xy, xz, yx, yy, yz, zx, zy, zz]

        If ``mode`` is ``"eigs"``, the data will be returned as::

            [eig_1, eig_2, eig_3]

        In all cases these are collected in a results dictionary with keys
        corresponding to the selected properties, e.g.::

            {'absorption': [absorption], 'eps_real': [eps_real]}
    """
    energies = np.array(dielectric[0])

    # Work with eps as complex numbers in Nx3x3 matrix
    # First interpret 6-column data as symmetric matrix
    # Input form xx yy zz xy yz xz
    # Indices     0  1  2  3  4  5
    real_eps = np.array(dielectric[1])[:, [[0, 3, 5], [3, 1, 4], [5, 4, 2]]]
    imag_eps = np.array(dielectric[2])[:, [[0, 3, 5], [3, 1, 4], [5, 4, 2]]]
    eps_full = real_eps + 1j * imag_eps

    # take sqrt of eps matrix; if eps = V S V^-1; then eps^1/2 = V S^{1/2} V^-1;
    eigvals, eigvecs = np.linalg.eig(eps_full)

    # fancy einsum to calculate V S^{1/2} V^-1 at every energy
    n = np.einsum("ijk,ik,ikl->ijl", eigvecs, np.sqrt(eigvals), np.linalg.inv(eigvecs))

    # calculate optical absorption
    alpha = n.imag * energies[:, None, None] * 4 * np.pi / 1.23984212e-4

    # Invert epsilon to obtain energy-loss function
    loss = -np.linalg.inv(eps_full).imag

    if mode == "average":
        eps = np.linalg.eigvals(eps_full).mean(axis=1)
        n = np.linalg.eigvals(n).mean(axis=1)
        loss = np.linalg.eigvalsh(loss).mean(axis=1)
        alpha = np.linalg.eigvalsh(alpha).mean(axis=1)
    elif mode == "eigs":
        eps = eigvals
        n = np.linalg.eigvals(n)
        loss = np.linalg.eigvalsh(loss)
        alpha = np.linalg.eigvalsh(alpha)
    elif mode == "full":
        eps = eps_full.reshape(-1, 9)
        n = n.reshape(-1, 9)
        loss = loss.reshape(-1, 9)
        alpha = alpha.reshape(-1, 9)
    elif mode == "trace":
        eps = np.diagonal(eps_full.T)
        n = np.diagonal(n.T)
        loss = np.diagonal(loss.T)
        alpha = np.diagonal(alpha.T)
    else:
        raise ValueError(
            f"Unsupported mode: {mode}. Options are 'average', 'eigs', 'full', 'trace'"
        )

    data = {
        "eps_real": eps.real,
        "eps_imag": eps.imag,
        "n_real": n.real,
        "n_imag": n.imag,
        "loss": loss,
        "absorption": alpha,
    }

    return energies, {k: data[k] for k in properties}


def write_files(abs_data, basename="absorption", prefix=None, directory=None):
    """Write the absorption or loss spectra to a file.

    Note that this function expects to receive an iterable series of spectra.

    Args:
        abs_data (tuple): Series (either :obj:`list` or :obj:`tuple`) of
            optical absorption or loss spectra. Each spectrum should be
            formatted as a :obj:`tuple` of :obj:`list` of :obj:`float`. If the
            data has been averaged, each spectrum should be::

                ([energies], [alpha])

            Else, if the data has not been averaged, each spectrum should be::

                ([energies], [alpha_xx, alpha_yy, alpha_zz]).

        prefix (:obj:`str`, optional): Prefix for file names.
        directory (:obj:`str`, optional): The directory in which to save files.
    """

    for i, absorption in enumerate(abs_data):
        num_txt = f"_{i + 1}" if len(abs_data) > 1 else ""
        prefix_txt = f"{prefix}_" if prefix else ""
        filename = prefix_txt + basename + num_txt + ".dat"

        if directory:
            filename = os.path.join(directory, filename)

        header = "energy(eV)"
        if len(absorption[1].shape) == 2:
            header += " alpha_xx alpha_yy alpha_zz"
            data = np.concatenate((absorption[0][:, None], absorption[1]), axis=1)
        else:
            header += " alpha"
            data = np.stack((absorption[0], absorption[1]), axis=1)

        np.savetxt(filename, data, header=header)


def kkr(de, eps_imag, cshift=1e-6):
    """Kramers Kronig transformation of imaginary dielectric function

    Args:
        de (:obj:`float`): Energy difference between evenly-spaced energy
            values corresponding to dielectric data
        eps_imag (:obj:`list` or :obj:`np.array`): Evenly-spaced sequence of
            frequency-dependent 3x3 dielectric matrices (imaginary component
            only)
        cshift (:obj:`float`, optional): imaginary finite shift used in
            integration; this should be small (and results should not be very
            sensitive)

    Returns:
        (:obj:`numpy.array`) Real part of frequency-dependent dielectric function
        corresponding to eps_imag. Array shape (NEDOS, 3, 3)
    """
    eps_imag = np.array(eps_imag)
    nedos = eps_imag.shape[0]
    cshift = complex(0, cshift)
    w_i = np.arange(0, (nedos - 0.5) * de, de, dtype=np.complex_)
    w_i = np.reshape(w_i, (nedos, 1, 1))

    def integration_element(w_r):
        factor = w_i / (w_i**2 - w_r**2 + cshift)
        total = np.sum(eps_imag * factor, axis=0)
        return total * (2 / np.pi) * de + np.diag([1, 1, 1])

    return np.real([integration_element(w_r) for w_r in w_i[:, 0, 0]])



# Copyright (c) Scanlon Materials Theory Group
# Distributed under the terms of the MIT License.

"""
Module containing class for generating k-points along paths from SeeK-path.
"""

from itertools import chain

from sumo.symmetry import Kpath


class SeekpathKpath(Kpath):
    r"""Class to generate k-points along paths from SeeK-path.

    More detail on the paths generated by SeeK-path can be found in
    reference [seek]_.

    .. [seek] Y. Hinuma, G. Pizzi, Y. Kumagai, F. Oba, I. Tanaka, Band
              structure diagram paths based on crystallography, Comp. Mat. Sci.
              128, 140 (2017). doi: 10.1016/j.commatsci.2016.10.015

    These paths should be used with primitive structures that comply with the
    definition from the paper. This structure can be accessed using the
    ``prim`` attribute and compliance between the provided structure and
    standardised structure checked using the ``correct_structure()`` method.

    Args:
        structure (:obj:`~pymatgen.core.structure.Structure`): The structure.
        symprec (:obj:`float`, optional): The tolerance for determining the
            crystal symmetry.

    Attributes:
        prim (:obj:`~pymatgen.core.structure.Structure`): The standardised
            primitive cell structure for the generated k-point path.
        conv (:obj:`~pymatgen.core.structure.Structure`): The standardised
            conventional cell structure.

    """

    def __init__(self, structure, symprec=1e-3):
        Kpath.__init__(self, structure, symprec=symprec)

        self._kpath = self.kpath_from_seekpath(
            self._seek_data["path"], self._seek_data["point_coords"]
        )

    @classmethod
    def kpath_from_seekpath(cls, seekpath, point_coords):
        r"""Convert seekpath-formatted kpoints path to sumo-preferred format.

        If 'GAMMA' is used as a label this will be replaced by '\Gamma'.

        Args:
            seekpath (list): A :obj:`list` of 2-tuples containing the labels at
                each side of each segment of the k-point path::

                    [(A, B), (B, C), (C, D), ...]

                where a break in the sequence is indicated by a non-repeating
                label. E.g.::

                    [(A, B), (B, C), (D, E), ...]

                for a break between C and D.
            point_coords (dict): Dict of coordinates corresponding to k-point
                labels::

                    {'GAMMA': [0., 0., 0.], ...}
        Returns:
            dict: The path and k-points as::

                {
                    'path', [[l1, l2, l3], [l4, l5], ...],
                    'kpoints', {l1: [a1, b1, c1], l2: [a2, b2, c2], ...}
                }
        """
        # convert from seekpath format e.g. [(l1, l2), (l2, l3), (l4, l5)]
        # to our preferred representation [[l1, l2, l3], [l4, l5]]
        path = [[seekpath[0][0]]]
        for (k1, k2) in seekpath:
            if path[-1] and path[-1][-1] == k1:
                path[-1].append(k2)
            else:
                path.append([k1, k2])

        # Rebuild kpoints dictionary skipping any positions not on path
        # (chain(*list) flattens nested list; set() removes duplicates.)
        kpoints = {p: point_coords[p] for p in set(chain(*path))}

        # Every path should include Gamma-point. Change the label to \Gamma
        assert "GAMMA" in kpoints
        kpoints[r"\Gamma"] = kpoints.pop("GAMMA")
        path = [
            [label.replace("GAMMA", r"\Gamma") for label in subpath] for subpath in path
        ]

        return {"kpoints": kpoints, "path": path}



import unittest
from os.path import join as path_join

import numpy as np
from pkg_resources import resource_filename
from pymatgen.core.lattice import Lattice

from sumo.io.questaal import QuestaalInit, dielectric_from_file


class QuestaalOpticsTestCase(unittest.TestCase):
    def setUp(self):
        self.bse_path = resource_filename(
            __name__, path_join("..", "data", "SnO2", "eps_BSE.out")
        )

    def test_optics_from_bethesalpeter(self):
        energy, real, imag = dielectric_from_file(self.bse_path)
        self.assertEqual(len(energy), 8000)
        self.assertAlmostEqual(energy[1], 3.401875234404301e-003)
        self.assertAlmostEqual(real[2][0], 1.94823351685956)
        self.assertAlmostEqual(imag[3][2], 1.084598600094848e-002)


class QuestaalInitTestCase(unittest.TestCase):
    def setUp(self):
        self.ref_lat = np.array(
            [[1.59205, -2.757511, 0.0], [1.59205, 2.757511, 0.0], [0.0, 0.0, 5.1551]]
        )

        self.ref_pmg_lat = Lattice(self.ref_lat)

    def test_init_from_python(self):
        """Check Questaal input object"""
        # spcgroup example
        lattice = {"SPCGRP": 186, "A": 3.18409958, "C": 5.1551, "UNITS": "A", "ALAT": 1}
        site = [
            {"ATOM": "Zn", "X": (0.6666670, 0.3333330, 0.5000000)},
            {"ATOM": "O", "X": (0.6666670, 0.3333330, 0.8803100)},
        ]

        init_sym = QuestaalInit(lattice, site)

        init_sym.structure
        self.assertFalse(init_sym.cartesian)

        # Check ALAT ok
        init_sym_alat = QuestaalInit(
            {"SPCGRP": 186, "A": 0.318409958, "C": 0.51551, "UNITS": "A", "ALAT": 10},
            site,
        )
        self.assertEqual(init_sym.structure, init_sym_alat.structure)

        # unitcell example
        lattice = {
            "ALAT": 1,
            "UNITS": "A",
            "PLAT": [
                [1.5920500, -2.7575110, 0.0000000],
                [1.5920500, 2.7575110, 0.0000000],
                [0.0000000, 0.0000000, 5.1551000],
            ],
        }
        site = [
            {"ATOM": "Zn", "X": (0.6666670, 0.3333330, 0.5000000)},
            {"ATOM": "Zn", "X": (0.3333330, 0.6666670, 1.0000000)},
            {"ATOM": "O", "X": (0.6666670, 0.3333330, 0.8803100)},
            {"ATOM": "O", "X": (0.3333330, 0.6666670, 0.3803100)},
        ]

        init_plat = QuestaalInit(lattice, site)

        init_plat.structure
        self.assertFalse(init_plat.cartesian)

        init_plat_alat = QuestaalInit(lattice, site)
        init_plat_alat.lattice["PLAT"] = np.array(init_plat_alat.lattice["PLAT"]) * 0.1
        init_plat_alat.lattice["ALAT"] = 10
        self.assertEqual(init_plat.structure, init_plat_alat.structure)

        # Bohr, Cartesian units
        bohr_lattice = {
            "ALAT": 1,
            "UNITS": None,
            "PLAT": [
                [3.00853848, -5.21094058, 0.0],
                [3.00853848, 5.21094058, 0.0],
                [0.0, 0.0, 9.74172715],
            ],
        }
        bohr_cart_sites = [
            {"ATOM": "Zn", "POS": (3.00853848, -1.73698367, 4.87086358)},
            {"ATOM": "Zn", "POS": (3.00853848, 1.73698367, 9.74172715)},
            {"ATOM": "O", "POS": (3.00853848, -1.73698367, 8.57573983)},
            {"ATOM": "O", "POS": (3.00853848, 1.73698367, 3.70487625)},
        ]

        bohr_init_lat = QuestaalInit(bohr_lattice, bohr_cart_sites)
        # Cartesian detected
        self.assertTrue(bohr_init_lat.cartesian)
        # unit conversion: lattice
        self.assertLess(
            (
                abs(
                    np.array(self.ref_pmg_lat.abc)
                    - np.array(bohr_init_lat.structure.lattice.abc)
                )
            ).max(),
            1e-5,
        )
        # unit conversion: sites  (Use distance matrix so translation ignored)
        self.assertLess(
            (
                abs(
                    init_plat.structure.distance_matrix
                    - bohr_init_lat.structure.distance_matrix
                )
            ).max(),
            1e-5,
        )
        # Ignore_units option
        bohr_init_noconvert = QuestaalInit(
            bohr_lattice, bohr_cart_sites, ignore_units=True
        )
        self.assertAlmostEqual(bohr_init_noconvert.structure.lattice.abc[2], 9.74172715)
        self.assertAlmostEqual(
            bohr_init_noconvert.structure.distance_matrix[0, -1], 3.66441077
        )

    def test_init_coordinate_safety(self):
        """Check illegal Questaal input caught"""
        lattice = {"SPCGRP": 186, "A": 3.18409958, "C": 5.1551, "UNITS": "A", "ALAT": 1}
        site = [
            {"ATOM": "Zn", "X": (0.6666670, 0.3333330, 0.5000000)},
            {"ATOM": "O", "POS": (0.6666670, 0.3333330, 0.8803100)},
        ]

        with self.assertRaises(ValueError):
            QuestaalInit(lattice, site)

        site = [
            {"ATOM": "Zn", "C": (0.6666670, 0.3333330, 0.5000000)},
            {"ATOM": "O", "C": (0.6666670, 0.3333330, 0.8803100)},
        ]
        with self.assertRaises(NotImplementedError):
            QuestaalInit(lattice, site)

    def test_init_from_file(self):
        zno_path = resource_filename(__name__, path_join("..", "data", "ZnO"))
        init1 = QuestaalInit.from_file(
            path_join(zno_path, "init.zno_nosym"), preprocessor=False
        )
        init2 = QuestaalInit.from_file(
            path_join(zno_path, "init.zno_sym"), preprocessor=False
        )
        nosym_structure = init1.structure
        sym_structure = init2.structure

        self.assertLess(
            (
                abs(
                    np.array(self.ref_pmg_lat.abc)
                    - np.array(nosym_structure.lattice.abc)
                )
            ).max(),
            1e-5,
        )
        self.assertLess(
            (
                abs(
                    np.array(self.ref_pmg_lat.angles)
                    - np.array(nosym_structure.lattice.angles)
                )
            ).max(),
            1e-3,
        )

        self.assertLess(
            (
                abs(
                    np.array(self.ref_pmg_lat.abc) - np.array(sym_structure.lattice.abc)
                )
            ).max(),
            1e-5,
        )
        self.assertLess(
            (
                abs(
                    np.array(self.ref_pmg_lat.angles)
                    - np.array(sym_structure.lattice.angles)
                )
            ).max(),
            1e-3,
        )



# Copyright (c) Scanlon Materials Theory Group
# Distributed under the terms of the MIT License.

"""
Module containing the base class for high-symmetry k-point path determination.
"""

import numpy as np
import seekpath
import spglib
from pymatgen.core.structure import Structure
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer


class Kpath:
    r"""Base class providing helper functions for generating k-point paths.

    This class should not be used directly. Instead, one of the
    :obj:`~sumo.symmetry.brad_crack_kpath.BradCrackKpath`,
    :obj:`~sumo.symmetry.seekpath_kpath.SeekpathKpath`, or
    :obj:`~sumo.symmetry.custom_kpath.CustomKpath`, subclasses should be used.

    The main use of this parent object is for standardisation across the
    differing k-point path generation classes.

    Args:
        structure (:obj:`~pymatgen.core.structure.Structure`): The structure.
        symprec (:obj:`float`, optional): The tolerance for determining the
            crystal symmetry.

    Attributes:
        prim (:obj:`~pymatgen.core.structure.Structure`): The standardised
            primitive cell structure for the generated k-point path.
        conv (:obj:`~pymatgen.core.structure.Structure`): The standardised
            conventional cell structure.
    """

    def __init__(self, structure, symprec=1e-3):
        self.structure = structure

        # use sym as a quick way to access the cell data
        sym = SpacegroupAnalyzer(structure, symprec=symprec)
        self._spg_data = sym.get_symmetry_dataset()

        # make primitive and conventional cell from seekpath output
        std = spglib.refine_cell(sym._cell, symprec=symprec)
        self._seek_data = seekpath.get_path(std)

        prim_lattice = self._seek_data["primitive_lattice"]
        prim_scaled_positions = self._seek_data["primitive_positions"]
        prim_numbers = self._seek_data["primitive_types"]
        prim_atoms = [sym._unique_species[i - 1] for i in prim_numbers]
        self.prim = Structure(prim_lattice, prim_atoms, prim_scaled_positions)

        conv_lattice = self._seek_data["conv_lattice"]
        conv_scaled_positions = self._seek_data["conv_positions"]
        conv_numbers = self._seek_data["conv_types"]
        conv_atoms = [sym._unique_species[i - 1] for i in conv_numbers]
        self.conv = Structure(conv_lattice, conv_atoms, conv_scaled_positions)

    def correct_structure(self, atol=1e-8):
        """Determine if the structure matches the standard primitive structure.

        The standard primitive will be different between seekpath and pymatgen
        high-symmetry paths, but this is handled by the specific subclasses.

        Args:
            atol (:obj:`float`, optional): Absolute tolerance used to compare
                the input structure with the primitive standard structure.

        Returns:
            bool: ``True`` if the structure is the same as the standard
            primitive, otherwise ``False``.
        """
        return np.allclose(
            self.structure.lattice.matrix, self.prim.lattice.matrix, atol=atol
        )

    def get_kpoints(self, line_density=20, cart_coords=False, phonopy=False):
        r"""Return a list of k-points and labels along the high-symmetry path.

        The format of the returned data will be different if phonopy is
        ``True`` or ``False``. This is because phonopy requires the labels and
        kpoints to be provided in a different format than kgen.

        Adapted from
        :obj:`pymatgen.symmetry.bandstructure.HighSymmKpath.get_kpoints`.

        Args:
            line_density (:obj:`int`, optional): Density of k-points along the
                path.
            cart_coords (:obj:`bool`, optional): Whether the k-points are
                returned in cartesian or reciprocal coordinates. Defaults to
                ``False`` (fractional coordinates).
            phonopy (:obj:`bool`, optional): Format the k-points and labels for
                use with phonopy. Defaults to ``False``.

        Returns:
            tuple: A :obj:`tuple` of the k-points along the high-symmetry path,
            and k-point labels. Returned as ``(kpoints, labels)``.

            If ``phonopy == False``, then:

                * ``kpoints`` is a :obj:`numpy.ndarray` of the k-point
                  coordinates along the high-symmetry path. For example::

                      [[0, 0, 0], [0.25, 0, 0], [0.5, 0, 0], [0.5, 0, 0.25],
                       [0.5, 0, 0.5]]

                * ``labels`` is a :obj:`list` of the high symmetry labels for
                  each k-point (will be an empty :obj:`str` if the k-point has
                  no label). For example::

                      ['\Gamma', '', 'X', '', 'Y']

            If ``phonopy == True``, then:

                * ``kpoints`` is a :obj:`list` of :obj:`numpy.ndarray`
                  containing the k-points for each branch of the band
                  structure. This means that the first and last k-points of a
                  particular branch may be repeated. For example::

                      [[[0, 0, 0], [0.25, 0, 0], [0.5, 0, 0]],
                       [[0.5, 0, 0], [0.5, 0, 0.25], [0.5, 0, 0.5]]]

                * ``labels`` is a :obj:`list` of the high symmetry labels.
                  For example::

                      ['\Gamma', 'X', 'Y']
        """
        list_k_points = []
        sym_point_labels = []
        recip_lattice = self.structure.lattice.reciprocal_lattice
        for b in self.path:
            for i in range(1, len(b)):
                start = np.array(self.kpoints[b[i - 1]])
                end = np.array(self.kpoints[b[i]])
                distance = np.linalg.norm(
                    recip_lattice.get_cartesian_coords(start)
                    - recip_lattice.get_cartesian_coords(end)
                )
                nb = int(np.ceil(distance * line_density))
                sym_point_labels.extend([b[i - 1]] + [""] * (nb - 1))

                limit = nb + 1 if phonopy else nb
                kpts = [
                    recip_lattice.get_cartesian_coords(start)
                    + float(i)
                    / float(nb)
                    * (
                        recip_lattice.get_cartesian_coords(end)
                        - recip_lattice.get_cartesian_coords(start)
                    )
                    for i in range(0, limit)
                ]

                if phonopy:
                    list_k_points.append(kpts)
                else:
                    list_k_points.extend(kpts)

            # append last k-point to avoid repetition as in pymatgen
            if not phonopy:
                # for VASP we label every k-point. If a k-point has no
                # high-symmetry label then just use an empty string.
                sym_point_labels.append(b[-1])
                list_k_points.append(recip_lattice.get_cartesian_coords(end))

        if phonopy:
            # For phonopy, the labels for any discontinuities should be
            # combined. For example if the route is  X -> Y | Z -> R, the path
            # will be [['X', 'Y'], ['Z', 'R']], and the labels should be
            # ['X', 'Z', 'R']

            sym_point_labels = []
            for i, path_branch in enumerate(self.path):
                for n, label in enumerate(path_branch):
                    if i != 0 and n == 0:
                        sym_point_labels[-1] += f" | {label}"
                    else:
                        sym_point_labels.append(label)

        if cart_coords:
            return list_k_points, sym_point_labels
        else:
            if phonopy:
                frac_k_points = [
                    [recip_lattice.get_fractional_coords(k) for k in p]
                    for p in list_k_points
                ]
                frac_k_points = frac_k_points
            else:
                frac_k_points = [
                    recip_lattice.get_fractional_coords(k) for k in list_k_points
                ]
            return frac_k_points, sym_point_labels

    @property
    def kpoints(self):
        r"""dict: The high-symmetry k-point labels and their fractional
        coordinates. Formatted as ``{label: coords}``. For example::

            {'\Gamma': [0., 0., 0.], 'X': [0.5, 0. 0.]}
        """
        return self._kpath["kpoints"]

    @property
    def path(self):
        r"""list: The high-symmetry k-point path. Each subpath is provided
        as a list. For example, the following covers the path ``\Gamma ->
        X -> C | \Gamma -> Y``::

            [['\Gamma', 'X', 'C'], ['\Gamma', 'Y']].
        """
        return self._kpath["path"]

    @property
    def lattice_type(self):
        """str: The Bravais lattice system. Hexagonal cells are separated into
        rhombohedral and hexagonal lattices.
        """
        return self.get_lattice_type(self.spg_number)

    @property
    def spg_symbol(self):
        """str: The international space group symbol."""
        return self._spg_data["international"]

    @property
    def spg_number(self):
        """int: The international space group number."""
        return self._spg_data["number"]

    @property
    def path_string(self):
        """str: The high-symmetry k-point path formatted with arrows and
        showing disconnections between subpaths. For example::

            "X -> Gamma | Y -> Z"
        """
        return " | ".join([" -> ".join(subpath) for subpath in self.path])

    @staticmethod
    def get_lattice_type(number):
        """Return the lattice crystal system.

        Hexagonal cells are differentiated into rhombohedral and hexagonal
        lattices.

        Args:
            number (int): The international space group number.

        Returns:
            str: The lattice crystal system.
        """

        def f(i, j):
            return i <= number <= j

        cs = {
            "triclinic": (1, 2),
            "monoclinic": (3, 15),
            "orthorhombic": (16, 74),
            "tetragonal": (75, 142),
            "trigonal": (143, 167),
            "hexagonal": (168, 194),
            "cubic": (195, 230),
        }

        crystal_system = None
        for k, v in cs.items():
            if f(*v):
                crystal_system = k
                break

        if number in [146, 148, 155, 160, 161, 166, 167]:
            return "rhombohedral"
        elif crystal_system == "trigonal":
            return "hexagonal"
        else:
            return crystal_system



# Copyright (c) Scanlon Materials Theory Group
# Distributed under the terms of the MIT License.

"""
Subpackage providing helper functions for generating publication ready plots.
"""
from functools import wraps

import matplotlib.pyplot
import numpy as np
from matplotlib import rcParams
from matplotlib.collections import LineCollection
from pkg_resources import resource_filename

colour_cache = {}

sumo_base_style = resource_filename("sumo.plotting", "sumo_base.mplstyle")
sumo_dos_style = resource_filename("sumo.plotting", "sumo_dos.mplstyle")
sumo_bs_style = resource_filename("sumo.plotting", "sumo_bs.mplstyle")
sumo_phonon_style = resource_filename("sumo.plotting", "sumo_phonon.mplstyle")
sumo_optics_style = resource_filename("sumo.plotting", "sumo_optics.mplstyle")


def styled_plot(*style_sheets):
    """Return a decorator that will apply matplotlib style sheets to a plot.

    ``style_sheets`` is a base set of styles, which will be ignored if
    ``no_base_style`` is set in the decorated function arguments.

    The style will further be overwritten by any styles in the ``style``
    optional argument of the decorated function.

    Args:
        style_sheets (:obj:`list`, :obj:`str`, or :obj:`dict`): Any matplotlib
            supported definition of a style sheet. Can be a list of style of
            style sheets.
    """

    def decorator(get_plot):
        @wraps(get_plot)
        def wrapper(*args, fonts=None, style=None, no_base_style=False, **kwargs):

            if no_base_style:
                list_style = []
            else:
                list_style = list(style_sheets)

            if style is not None:
                if isinstance(style, list):
                    list_style += style
                else:
                    list_style += [style]

            if fonts is not None:
                list_style += [{"font.family": "sans-serif", "font.sans-serif": fonts}]

            matplotlib.pyplot.style.use(list_style)
            return get_plot(*args, **kwargs)

        return wrapper

    return decorator


def pretty_plot(width=None, height=None, plt=None, dpi=None):
    """Get a :obj:`matplotlib.pyplot` object with publication ready defaults.

    Args:
        width (:obj:`float`, optional): The width of the plot.
        height (:obj:`float`, optional): The height of the plot.
        plt (:obj:`matplotlib.pyplot`, optional): A :obj:`matplotlib.pyplot`
            object to use for plotting.
        dpi (:obj:`int`, optional): The dots-per-inch (pixel density) for
            the plot.

    Returns:
        :obj:`matplotlib.pyplot`: A :obj:`matplotlib.pyplot` object with
        publication ready defaults set.
    """

    if plt is None:
        plt = matplotlib.pyplot
        if width is None:
            width = matplotlib.rcParams["figure.figsize"][0]
        if height is None:
            height = matplotlib.rcParams["figure.figsize"][1]

        if dpi is not None:
            matplotlib.rcParams["figure.dpi"] = dpi

        fig = plt.figure(figsize=(width, height))
        fig.add_subplot(1, 1, 1)

    return plt


def pretty_subplot(
    nrows,
    ncols,
    width=None,
    height=None,
    sharex=True,
    sharey=True,
    dpi=None,
    plt=None,
    gridspec_kw=None,
):
    """Get a :obj:`matplotlib.pyplot` subplot object with pretty defaults.

    Args:
        nrows (int): The number of rows in the subplot.
        ncols (int): The number of columns in the subplot.
        width (:obj:`float`, optional): The width of the plot.
        height (:obj:`float`, optional): The height of the plot.
        sharex (:obj:`bool`, optional): All subplots share the same x-axis.
            Defaults to ``True``.
        sharey (:obj:`bool`, optional): All subplots share the same y-axis.
            Defaults to ``True``.
        dpi (:obj:`int`, optional): The dots-per-inch (pixel density) for
            the plot.
        plt (:obj:`matplotlib.pyplot`, optional): A :obj:`matplotlib.pyplot`
            object to use for plotting.
        gridspec_kw (:obj:`dict`, optional): Gridspec parameters. Please see:
            :obj:`matplotlib.pyplot.subplot` for more information. Defaults
            to ``None``.

    Returns:
        :obj:`matplotlib.pyplot`: A :obj:`matplotlib.pyplot` subplot object
        with publication ready defaults set.
    """

    if width is None:
        width = rcParams["figure.figsize"][0]
    if height is None:
        height = rcParams["figure.figsize"][1]

    # TODO: Make this work if plt is already set...
    if plt is None:
        plt = matplotlib.pyplot
        plt.subplots(
            nrows,
            ncols,
            sharex=sharex,
            sharey=sharey,
            dpi=dpi,
            figsize=(width, height),
            facecolor="w",
            gridspec_kw=gridspec_kw,
        )

    return plt


def curry_power_tick(times_sign=r"\times"):
    def f(val, pos):
        return power_tick(val, pos, times_sign=times_sign)

    return f


def power_tick(val, pos, times_sign=r"\times"):
    """Custom power ticker function."""
    if val == 0:
        return r"$\mathregular{0}$"
    elif val < 0:
        exponent = int(np.log10(-val))
    else:
        exponent = int(np.log10(val))
    coeff = val / 10**exponent
    prec = 0 if coeff % 1 == 0 else 1

    return rf"$\mathregular{{{coeff:.{prec}f} {times_sign} 10^{exponent:2d}}}$"


def colorline(
    x,
    y,
    weights,
    color1="#FF0000",
    color2="#00FF00",
    color3="#0000FF",
    colorspace="lab",
    linestyles="solid",
    linewidth=2.5,
):
    """Get a RGB coloured line for plotting.

    Args:
        x (list): x-axis data.
        y (list): y-axis data (can be multidimensional array).
        weights (list): The weights of the color1, color2, and color3 channels. Given
            as an array with the shape (n, 3), where n is the same length as the x and
            y data.
        color1 (str): A color specified in any way supported by matplotlib.
        color2 (str): A color specified in any way supported by matplotlib.
        color3 (str): A color specified in any way supported by matplotlib.
        colorspace (str): The colorspace in which to perform the interpolation. The
            allowed values are rgb, hsv, lab, luvlc, lablch, and xyz.
        linestyles (:obj:`str`, optional): Linestyle for plot. Options are
            ``"solid"`` or ``"dotted"``.
    """
    y = np.array(y)
    if len(y.shape) == 1:
        y = np.array([y])
        weights = np.array([weights])

    seg = []
    colours = []
    for yy, ww in zip(y, weights):
        pts = np.array([x, yy]).T.reshape(-1, 1, 2)
        if len(pts) > 1:  # need at least one point to interpolate colours
            seg.extend(np.concatenate([pts[:-1], pts[1:]], axis=1))

            nseg = len(x) - 1
            w = [0.5 * (ww[i] + ww[i + 1]) for i in range(nseg)]
            c = get_interpolated_colors(
                color1, color2, color3, w, colorspace=colorspace
            )
            colours.extend(c.tolist())

    lc = LineCollection(
        seg, colors=colours, rasterized=True, linewidth=linewidth, linestyles=linestyles
    )
    return lc


def get_interpolated_colors(color1, color2, color3, weights, colorspace="lab"):
    """
    Interpolate colors at a number of points within a colorspace.

    Args:
        color1 (str): A color specified in any way supported by matplotlib.
        color2 (str): A color specified in any way supported by matplotlib.
        color3 (str): A color specified in any way supported by matplotlib.
        weights (list): A list of weights with the shape (n, 3). Where the 3 values of
            the last axis give the amount of color1, color2, and color3.
        colorspace (str): The colorspace in which to perform the interpolation. The
            allowed values are rgb, hsv, lab, luvlc, lablch, and xyz.

    Returns:
        A list of colors, specified in the rgb format as a (n, 3) array.
    """
    from colormath.color_conversions import convert_color
    from colormath.color_objects import (
        HSVColor,
        LabColor,
        LCHabColor,
        LCHuvColor,
        XYZColor,
        sRGBColor,
    )
    from matplotlib.colors import to_rgb

    colorspace_mapping = {
        "rgb": sRGBColor,
        "hsv": HSVColor,
        "lab": LabColor,
        "luvlch": LCHuvColor,
        "lablch": LCHabColor,
        "xyz": XYZColor,
    }
    if colorspace not in list(colorspace_mapping.keys()):
        raise ValueError(f"colorspace must be one of {colorspace_mapping.keys()}")

    colorspace = colorspace_mapping[colorspace]

    # first convert matplotlib color specification to colormath sRGB
    color1_rgb = sRGBColor(*to_rgb(color1))
    color2_rgb = sRGBColor(*to_rgb(color2))
    color3_rgb = sRGBColor(*to_rgb(color3))

    # now convert to the colorspace basis for interpolation
    basis1 = np.array(
        convert_color(color1_rgb, colorspace, target_illuminant="d50").get_value_tuple()
    )
    basis2 = np.array(
        convert_color(color2_rgb, colorspace, target_illuminant="d50").get_value_tuple()
    )
    basis3 = np.array(
        convert_color(color3_rgb, colorspace, target_illuminant="d50").get_value_tuple()
    )

    # ensure weights is a numpy array
    weights = np.asarray(weights)

    # perform the interpolation in the colorspace basis
    colors = (
        basis1 * weights[:, 0][:, None]
        + basis2 * weights[:, 1][:, None]
        + basis3 * weights[:, 2][:, None]
    )

    # convert colors to RGB
    rgb_colors = [
        convert_color(colorspace(*c), sRGBColor).get_value_tuple() for c in colors
    ]

    # ensure all rgb values are less than 1 (sometimes issues in interpolation gives
    # values slightly over 1)
    return np.minimum(rgb_colors, 1)


def draw_themed_line(y, ax, orientation="horizontal"):
    """Draw a horizontal line using the theme settings

    Args:
        y (float): Position of line in data coordinates
        ax (Axes): Matplotlib Axes on which line is drawn

    """

    # Note to future developers: feel free to add plenty more optional
    # arguments to this to mess with linestyle, zorder etc.
    # Just .update() the options dict

    themed_line_options = dict(
        color=rcParams["grid.color"],
        linestyle="--",
        dashes=(5, 2),
        zorder=0,
        linewidth=rcParams["ytick.major.width"],
    )

    if orientation == "horizontal":
        ax.axhline(y, **themed_line_options)
    elif orientation == "vertical":
        ax.axvline(y, **themed_line_options)
    else:
        raise ValueError(f'Line orientation "{orientation}" not supported')



# Copyright (c) Scanlon Materials Theory Group
# Distributed under the terms of the MIT License.

"""
Class for generating k-points along Latimer & Munro high-symmetry paths.
"""

from pymatgen.symmetry.bandstructure import HighSymmKpath

from sumo.symmetry import Kpath


class LatimerKpath(Kpath):
    r"""Class to generate k-points along paths from the Latimer & Munro method.

    More details on the paths generated by can be found in the pymatgen
    documentation. These paths are based on the work described in reference
    [lm]_.

    .. [lm] Munro, J.M., Latimer, K., Horton, M.K. et al.
        An improved symmetry-based approach to reciprocal space path selection
        in band structure calculations. npj Comput Mater 6, 112 (2020).
        doi: 10.1038/s41524-020-00383-7

    Args:
        structure (:obj:`~pymatgen.core.structure.Structure`): The structure.
        symprec (:obj:`float`, optional): The tolerance for determining the
            crystal symmetry.
    """

    def __init__(self, structure, symprec=1e-3):
        Kpath.__init__(self, structure, symprec=symprec)
        pmg_path = HighSymmKpath(self.prim, symprec=symprec, path_type="lm")
        self._kpath = pmg_path._kpath



# Copyright (c) Scanlon Materials Theory Group
# Distributed under the terms of the MIT License.

"""
This module provides a class for plotting optical absorption spectra.
"""

import numpy as np
import scipy.constants as scpc
from matplotlib import rcParams
from matplotlib.font_manager import FontProperties, findfont
from matplotlib.ticker import AutoMinorLocator, FuncFormatter, MaxNLocator

from sumo.plotting import (
    curry_power_tick,
    pretty_subplot,
    styled_plot,
    sumo_base_style,
    sumo_optics_style,
)


def ev_to_nm(energy_ev):  # convert energy in eV to wavelength in nm
    return 1e9 * (scpc.h * scpc.c) / (energy_ev * scpc.electron_volt)


class SOpticsPlotter:
    """Class for plotting optical spectra.


    The easiest way to initialise this class is using the output from the
    :obj:`sumo.electronic_structure.optics.calculate_alpha()` method.

    Args:
        spec_data (:obj:`dict` or :obj:`tuple` or :obj:`list`):
            The optical absorption spectra. Should be formatted as a dict of
            :obj:`tuple` of :obj:`list`::

                {property: ([energies], [alpha]), ...}

            Alternatively, the anisotropic (directional dependent) absorption
            can be plotted if the data formatted as::

                {property: ([energies], [alpha_xx, alpha_yy, alpha_zz]), ...}

            If a :obj:`list` of :obj:`tuple` is provided, then multiple
            absorption spectra can be plotted simultaneously.

            The *property* keys set the type of spectrum being plotted (and
            determine the y-axis label). Recognised values are 'absorption',
            'loss', 'eps-real', 'eps-im'. Other values will be plotted and used
            as the y-axis label.

            It is recommended to use an OrderedDict to get a predictable
            sequence of plotting axes (from top to bottom).

            If no dict at all is used (i.e. spec_data is a data tuple or a list
            of tuples) property will default to 'abs'

        band_gap (:obj:`float` or :obj:`list`, optional): The band gap as a
            :obj:`float`, plotted as a dashed line. If plotting multiple
            spectra then a :obj:`list` of band gaps can be provided. If an item
            in the list is :obj:`None` no corresponding line is plotted.
        label (:obj:`str` or :obj:`list`): A label to identify the spectra.
            If plotting multiple spectra then a :obj:`list` of labels can
            be provided.
    """

    def __init__(self, spec_data, band_gap=None, label=None):
        # Standardise to dict/list/tuple format
        if isinstance(spec_data, tuple):
            spec_data = {"absorption": [spec_data]}
        elif isinstance(spec_data, list):
            spec_data = {"absorption": spec_data}
        for k, v in spec_data.items():
            spec_data[k] = v if isinstance(v, list) else [v]

        n_datasets = len(next(iter(spec_data.items()))[1])

        if isinstance(band_gap, float):
            band_gap = [band_gap]
        elif not band_gap:
            band_gap = [None] * n_datasets

        if isinstance(label, str):
            label = [label]
        elif not label and n_datasets > 1:
            label = [str(i) for i in range(1, n_datasets + 1)]
        elif not label:
            label = [""] * n_datasets

        if len({len(label), n_datasets, len(band_gap)}) > 1:
            raise ValueError("spec_data, band_gap, and label not same size")

        if "absorption" in spec_data:
            # xmax: find lowest energy where absorption > 2e4; add 1 eV
            xmax = 0
            for ener, alpha in spec_data["absorption"]:
                mask = alpha > 2e4
                if len(alpha.shape) == 1:
                    x = min(ener[mask])
                else:
                    x = max(min(ener[mask[:, i]]) for i in range(3))
                xmax = x if x > xmax else xmax
        else:
            # If no absorption data, use the maximum of the shortest x range
            xmax = float("Inf")
            for _, spec in spec_data.items():
                for dataset in spec:
                    xmax = min((xmax, np.max(dataset[0])))

        self._spec_data = spec_data
        self._band_gap = band_gap
        self._label = label
        self._xmax = xmax + 1.0

    @styled_plot(sumo_base_style, sumo_optics_style)
    def get_plot(
        self,
        units="eV",
        width=None,
        height=None,
        xmin=0.0,
        xmax=None,
        ymin=0,
        ymax=None,
        colours=None,
        dpi=400,
        plt=None,
        fonts=None,
        style=None,
        no_base_style=False,
    ):
        """Get a :obj:`matplotlib.pyplot` object of the optical spectra.

        Args:
            units (:obj:`str`, optional): X-axis units for the plot. 'eV' for
                energy in electronvolts or 'nm' for wavelength in nanometers.
                Defaults to 'eV'.
            width (:obj:`float`, optional): The width of the plot.
            height (:obj:`float`, optional): The height of the plot.
            xmin (:obj:`float`, optional): The minimum energy on the x-axis.
            xmax (:obj:`float`, optional): The maximum energy on the x-axis.
            ymin (:obj:`float`, optional): The minimum absorption intensity on
                the y-axis.
            ymax (:obj:`float`, optional): The maximum absorption intensity on
                the y-axis.
            colours (:obj:`list`, optional): A :obj:`list` of colours to use in
                the plot. The colours can be specified as a hex code, set of
                rgb values, or any other format supported by matplotlib.
            dpi (:obj:`int`, optional): The dots-per-inch (pixel density) for
                the image.
            plt (:obj:`matplotlib.pyplot`, optional): A
                :obj:`matplotlib.pyplot` object to use for plotting.
            fonts (:obj:`list`, optional): Fonts to use in the plot. Can be a
                a single font, specified as a :obj:`str`, or several fonts,
                specified as a :obj:`list` of :obj:`str`.
            style (:obj:`list`, :obj:`str`, or :obj:`dict`): Any matplotlib
                style specifications, to be composed on top of Sumo base
                style.
            no_base_style (:obj:`bool`, optional): Prevent use of sumo base
                style. This can make alternative styles behave more
                predictably.

        Returns:
            :obj:`matplotlib.pyplot`: The plot of optical spectra.
        """
        n_plots = len(self._spec_data)
        plt = pretty_subplot(
            n_plots,
            1,
            sharex=True,
            sharey=False,
            width=width,
            height=height,
            dpi=dpi,
            plt=plt,
        )
        fig = plt.gcf()

        optics_colours = rcParams["axes.prop_cycle"].by_key()["color"]
        if colours is not None:
            optics_colours = colours + optics_colours

        standard_ylabels = {
            "absorption": r"Absorption (cm$^\mathregular{-1}$)",
            "loss": r"Energy-loss",
            "eps_real": r"Re($\epsilon$)",
            "eps_imag": r"Im($\epsilon$)",
            "n_real": r"Re(n)",
            "n_imag": r"Im(n)",
        }

        if ymax is None:
            ymax_series = [None] * n_plots
        elif isinstance(ymax, float) or isinstance(ymax, int):
            ymax_series = [ymax] * n_plots
        elif not isinstance(ymax, list):
            raise ValueError()
        else:
            ymax_series = ymax

        if ymin is None:
            ymin_series = [None] * n_plots
        elif isinstance(ymin, float) or isinstance(ymin, int):
            ymin_series = [ymin] * n_plots
        elif not isinstance(ymin, list):
            raise ValueError()
        else:
            ymin_series = ymin

        for i, (spectrum_key, data), ymin, ymax in zip(
            range(n_plots), self._spec_data.items(), ymin_series, ymax_series
        ):
            ax = fig.axes[i]
            _plot_spectrum(data, self._label, self._band_gap, ax, optics_colours, units)

            if units in ["ev", "eV"] and xmax is None:
                xmax = self._xmax  # use sumo-determined energy limits
            elif units == "nm":
                # use default minimum energy (max wavelength) of 2500 nm
                xmax = xmax if xmax is not None else 2500
                # convert sumo-determined max energy to min wavelength
                xmin = xmin if xmin else ev_to_nm(self._xmax)
            ax.set_xlim(xmin, xmax)

            if ymin is None and spectrum_key in (
                "absorption",
                "loss",
                "eps_imag",
                "n_imag",
            ):
                ymin = 0
            elif ymin is None:
                ymin = ax.get_ylim()[0]

            if ymax is None and spectrum_key in ("absorption",):
                ymax = 1e5
            elif ymax is None:
                ymax = ax.get_ylim()[1]

            ax.set_ylim(ymin, ymax)

            if spectrum_key == "absorption":
                font = findfont(FontProperties(family=["sans-serif"]))
                if "Whitney" in font:
                    times_sign = "x"
                else:
                    times_sign = r"\times"
                ax.yaxis.set_major_formatter(
                    FuncFormatter(curry_power_tick(times_sign=times_sign))
                )

            ax.yaxis.set_major_locator(MaxNLocator(5))
            ax.xaxis.set_major_locator(MaxNLocator(3))
            ax.yaxis.set_minor_locator(AutoMinorLocator(2))
            ax.xaxis.set_minor_locator(AutoMinorLocator(2))

            ax.set_ylabel(standard_ylabels.get(spectrum_key, spectrum_key))

            if i == 0:
                if (
                    not np.all(np.array(self._label) == "")
                    or len(np.array(next(iter(self._spec_data.items()))[1][0][1]).shape)
                    > 1
                ):
                    ax.legend(loc="best", frameon=False, ncol=1)

        xlabel = "Energy (eV)" if units == "eV" else "Wavelength (nm)"
        ax.set_xlabel(xlabel)

        # If only one plot, fix aspect ratio to match canvas
        if len(self._spec_data) == 1:
            x0, x1 = ax.get_xlim()
            y0, y1 = ax.get_ylim()
            if width is None:
                width = rcParams["figure.figsize"][0]
            if height is None:
                height = rcParams["figure.figsize"][1]
            ax.set_aspect((height / width) * ((x1 - x0) / (y1 - y0)))

        # Otherwise, rely only on tight_layout and hope for the best
        plt.tight_layout()
        return plt


def _plot_spectrum(data, label, band_gap, ax, optics_colours, units):
    for (ener, alpha), abs_label, bg, c in zip(data, label, band_gap, optics_colours):
        if len(alpha.shape) == 1:
            # if averaged optics only plot one line
            if units == "eV":
                ax.plot(ener, alpha, label=abs_label, c=c)
            elif (
                units == "nm"
            ):  # only for energies greater than 1 meV, to avoid RuntimeWarnings
                # when converting to nm
                ax.plot(
                    ev_to_nm(ener[ener > 0.001]),
                    alpha[ener > 0.001],
                    label=abs_label,
                    c=c,
                )

        else:
            data = zip(range(3), ["xx", "yy", "zz"], ["-", "--", "-."])

            for direction_id, direction_label, ls in data:
                if not abs_label:
                    label = direction_label
                else:
                    label = r"{}$_\mathregular{{{}}}$"
                    label = label.format(abs_label, direction_label)

                if units == "eV":
                    ax.plot(ener, alpha[:, direction_id], ls=ls, label=label, c=c)
                elif units == "nm":  # only for energies greater than 1 meV,
                    # to avoid RuntimeWarnings when converting to nm
                    ax.plot(
                        ev_to_nm(ener[ener > 0.001]),
                        alpha[ener > 0.001, direction_id],
                        ls=ls,
                        label=label,
                        c=c,
                    )
        if bg:
            # plot band gap line
            if units == "eV":
                ax.axvline(bg, ls=":", c=c)
            elif units == "nm":
                ax.axvline(ev_to_nm(bg), ls=":", c=c)



# Copyright (c) Scanlon Materials Theory Group
# Distributed under the terms of the MIT License.

"""
Module containing helper functions for calculating band effective masses.
"""

import numpy as np
from scipy.constants import physical_constants
from scipy.optimize import curve_fit

eV_to_hartree = physical_constants["electron volt-hartree relationship"][0]
bohr_to_m = physical_constants["Bohr radius"][0]
angstrom_to_bohr = bohr_to_m / 1e-10


def get_fitting_data(bs, spin, band_id, kpoint_id, num_sample_points=3):
    """Extract fitting data for band extrema based on spin, kpoint and band.

    Searches forward and backward from the extrema point, but will only sample
    there data if there are enough points in that direction.

    Args:
        bs (:obj:`~pymatgen.electronic_structure.bandstructure.BandStructureSymmLine`):
            The band structure.
        spin (:obj:`~pymatgen.electronic_structure.core.Spin`): Which spin
            channel to sample.
        band_id (int): Index of the band to sample.
        kpoint_id (int): Index of the kpoint to sample.

    Returns:
        list: The data necessary to calculate the effective mass, along with
        some metadata. Formatted as a :obj:`list` of :obj:`dict`, each with the
        keys:

        'energies' (:obj:`numpy.ndarray`)
            Band eigenvalues in eV.

        'distances' (:obj:`numpy.ndarray`)
            Distances of the k-points in reciprocal space.

        'band_id' (:obj:`int`)
            The index of the band,

        'spin' (:obj:`~pymatgen.electronic_structure.core.Spin`)
            The spin channel

        'start_kpoint' (:obj:`int`)
            The index of the k-point at which the band extrema occurs

        'end_kpoint' (:obj:`int`)
            The k-point towards which the data has been sampled.
    """
    # branch data provides data about the start and end points
    # of specific band paths
    branch_data = [b for b in bs.get_branch(kpoint_id) if b["index"] == kpoint_id][0]
    start_kpoint = bs.kpoints[kpoint_id]

    fitting_data = []

    # check to see if there are enough points to sample from first
    # check in the forward direction
    if kpoint_id + num_sample_points <= branch_data["end_index"]:

        # calculate sampling limits
        start_id = kpoint_id
        end_id = kpoint_id + num_sample_points + 1

        energies = np.array(bs.bands[spin][band_id][start_id:end_id].copy())
        dists = np.array(bs.distance[start_id:end_id].copy())

        # normalise eigenvalues and distances to starting point
        energies -= bs.bands[spin][band_id][kpoint_id]
        dists -= bs.distance[kpoint_id]

        # symmetrise the data to make fitting more reliable
        energies = np.concatenate([energies[::-1], energies[1:]])
        dists = np.concatenate([-dists[::-1], dists[1:]])

        end_kpoint = bs.kpoints[branch_data["end_index"]]
        data = {
            "energies": energies,
            "distances": dists,
            "band_id": band_id,
            "spin": spin,
            "start_kpoint": start_kpoint,
            "end_kpoint": end_kpoint,
        }
        fitting_data.append(data)

    # check in the backward direction
    if kpoint_id - num_sample_points >= branch_data["start_index"]:

        # calculate sampling limits
        start_id = kpoint_id - num_sample_points
        end_id = kpoint_id + 1

        energies = bs.bands[spin][band_id][start_id:end_id].copy()
        dists = bs.distance[start_id:end_id].copy()

        # normalise eigenvalues and distances to starting point
        energies -= bs.bands[spin][band_id][kpoint_id]
        dists -= bs.distance[kpoint_id]

        # symmetrise the data to make fitting more reliable
        energies = np.concatenate([energies[:-1], energies[::-1]])
        dists = np.concatenate([dists[:-1], -dists[::-1]])

        end_kpoint = bs.kpoints[branch_data["start_index"]]
        data = {
            "energies": energies,
            "distances": dists,
            "band_id": band_id,
            "spin": spin,
            "start_kpoint": start_kpoint,
            "end_kpoint": end_kpoint,
        }
        fitting_data.append(data)

    return fitting_data


def fit_effective_mass(distances, energies, parabolic=True):
    """Fit the effective masses using either a parabolic or nonparabolic fit.

    Args:
        distances (:obj:`numpy.ndarray`): The x-distances between k-points in
            reciprocal Angstroms, normalised to the band extrema.
        energies (:obj:`numpy.ndarray`): The band eigenvalues normalised to the
            eigenvalue of the band extrema.
        parabolic (:obj:`bool`, optional): Use a parabolic fit of the band
            edges. If ``False`` then nonparabolic fitting will be attempted.
            Defaults to ``True``.

    Returns:
        float: The effective mass in units of electron rest mass, :math:`m_0`.
    """
    if parabolic:
        fit = np.polyfit(distances, energies, 2)
        c = 2 * fit[0]  # curvature therefore 2 * the exponent on the ^2 term

    else:
        # Use non parabolic description of the bands
        def f(x, alpha, d):
            top = np.sqrt(4 * alpha * d * x**2 + 1) - 1
            bot = 2 * alpha
            return top / bot

        # set boundaries for curve fitting: alpha > 1e-8
        # as alpha = 0 causes an error
        bounds = ((1e-8, -np.inf), (np.inf, np.inf))
        popt, _ = curve_fit(f, distances, energies, p0=[1.0, 1.0], bounds=bounds)
        c = 2 * popt[1]

    # coefficient is currently in eV/Angstrom^2/h_bar^2
    # want it in atomic units so Hartree/bohr^2/h_bar^2
    eff_mass = (angstrom_to_bohr**2 / eV_to_hartree) / c
    return eff_mass



import unittest
from os.path import abspath

from pkg_resources import Requirement, resource_filename

try:
    import configparser
except ImportError:
    import ConfigParser as configparser

import matplotlib.pyplot

import sumo.plotting
import sumo.plotting.dos_plotter
from sumo.plotting.dos_plotter import get_cached_colour


class GetColourTestCase(unittest.TestCase):
    def setUp(self):
        # Open default CLI colours config
        config_path = resource_filename(
            Requirement.parse("sumo"), "sumo/plotting/orbital_colours.conf"
        )

        self.config = configparser.ConfigParser()
        self.config.read(abspath(config_path))

    def test_get_colour_cache(self):
        """Check colour caching"""
        col1, cache = get_cached_colour("Hf", "s", cache={})
        col2, cache = get_cached_colour("Zr", "d", cache=cache)
        col3, cache = get_cached_colour("Hf", "s", cache=cache)
        self.assertEqual(col1, col3)
        self.assertNotEqual(col1, col2)

        # Try rebooting with new cache
        col4, cache = tuple(get_cached_colour("Zr", "d", cache={}))
        self.assertEqual(col1, col4)

    def test_get_colour_global_cache(self):
        """Check colour caching"""
        sumo.plotting.colour_cache.clear()
        col1, _ = get_cached_colour("Hf", "s")
        col2, _ = get_cached_colour("Zr", "d")
        col3, _ = get_cached_colour("Hf", "s")
        self.assertEqual(col1, col3)
        self.assertNotEqual(col1, col2)

        # Try rebooting with new cache
        sumo.plotting.colour_cache.clear()
        col4, _ = get_cached_colour("Zr", "d")
        self.assertEqual(col1, col4)

    def test_get_colour_type_error(self):
        """Check bogus colour info is rejected"""
        with self.assertRaises(TypeError):
            get_cached_colour("Na", "p", colours=("#aabbcc"))

    def test_get_colour_config(self):
        """Check orbital colours from config file"""
        col_O_p, _ = get_cached_colour("O", "p", colours=self.config)
        col_Re_d, _ = get_cached_colour("Re", "d", colours=self.config)

        self.assertEqual(col_O_p, "#0DB14B")
        self.assertEqual(col_Re_d, "#A154A1")

    def test_get_colour_mixed(self):
        """Check new colours drawn in correct sequence"""
        sumo.plotting.colour_cache.clear()
        with matplotlib.pyplot.style.context("ggplot"):
            _, __ = get_cached_colour("O", "p", colours=self.config)
            col_Hf_s, _ = get_cached_colour("Hf", "s", colours=self.config)
            _, __ = get_cached_colour("Re", "d", colours=self.config)
            col_Zr_d, _ = get_cached_colour("Zr", "d", colours=self.config)

            prop_cyc = matplotlib.rcParams["axes.prop_cycle"].by_key()["color"]
            default_0 = prop_cyc[0]
            default_1 = prop_cyc[1]
            self.assertEqual(col_Hf_s, default_0)
            self.assertEqual(col_Zr_d, default_1)


if __name__ == "__main__":
    unittest.main()



import unittest
from os.path import join as path_join

import numpy as np
from phonopy import Phonopy
from pkg_resources import resource_filename
from pymatgen.io.vasp.inputs import Poscar

import sumo.phonon.phonopy


class PhonopyTestCase(unittest.TestCase):
    def setUp(self):
        self.phonon_data = resource_filename(
            __name__, path_join("..", "data", "RbSnI6", "phonopy", "FORCE_SETS")
        )
        poscar_path = resource_filename(
            __name__, path_join("..", "data", "RbSnI6", "phonopy", "POSCAR")
        )
        phonon_poscar = Poscar.from_file(poscar_path)
        self.phonon_structure = phonon_poscar.structure

    def test_phonopy_load(self):
        phonopy_obj = sumo.phonon.phonopy.load_phonopy(
            self.phonon_data, self.phonon_structure, np.diag([3, 3, 2])
        )
        self.assertIsInstance(phonopy_obj, Phonopy)

        self.assertEqual(phonopy_obj.force_constants.shape, (324, 324, 3, 3))
        self.assertAlmostEqual(
            phonopy_obj.force_constants.trace().trace(), 2654.3381334806545
        )
        self.assertAlmostEqual(
            phonopy_obj.force_constants[4, 3, 2, 1], 0.000552387416908992
        )



# Copyright (c) Scanlon Materials Theory Group
# Distributed under the terms of the MIT License.

"""
A script to calculate effective masses and band degeneracies.

TODO:
 - Would be good to reimplement get_vbm and get_cbm methods, with the ability
   to give degeneracies + other band edges within a certain tolerance.
 - Sample custom k-point, band, spin combinations.
"""

import argparse
import logging
import sys

from pymatgen.electronic_structure.core import Spin
from pymatgen.io.vasp.outputs import BSVasprun

from sumo.cli.bandplot import find_vasprun_files
from sumo.electronic_structure.bandstructure import get_reconstructed_band_structure
from sumo.electronic_structure.effective_mass import (
    fit_effective_mass,
    get_fitting_data,
)

__author__ = "Alex Ganose"
__version__ = "1.0"
__maintainer__ = "Alex Ganose"
__email__ = "alexganose@googlemail.com"
__date__ = "March 4, 2018"


kpt_str = "[{k[0]:.2f}, {k[1]:.2f}, {k[2]:.2f}]"


def bandstats(
    filenames=None,
    num_sample_points=3,
    temperature=None,
    degeneracy_tol=1e-4,
    parabolic=True,
):
    """Calculate the effective masses of the bands of a semiconductor.

    Args:
        filenames (:obj:`str` or :obj:`list`, optional): Path to vasprun.xml
            or vasprun.xml.gz file. If no filenames are provided, the code
            will search for vasprun.xml or vasprun.xml.gz files in folders
            named 'split-0*'. Failing that, the code will look for a vasprun in
            the current directory. If a :obj:`list` of vasprun files is
            provided, these will be combined into a single band structure.
        num_sample_points (:obj:`int`, optional): Number of k-points to sample
            when fitting the effective masses.
        temperature (:obj:`int`, optional): Find band edges within kB * T of
            the valence band maximum and conduction band minimum. Not currently
            implemented.
        degeneracy_tol (:obj:`float`, optional): Tolerance for determining the
            degeneracy of the valence band maximum and conduction band minimum.
        parabolic (:obj:`bool`, optional): Use a parabolic fit of the band
            edges. If ``False`` then nonparabolic fitting will be attempted.
            Defaults to ``True``.

    Returns:
        dict: The hole and electron effective masses. Formatted as a
        :obj:`dict` with keys: ``'hole_data'`` and ``'electron_data'``. The
        data is a :obj:`list` of :obj:`dict` with the keys:

        'effective_mass' (:obj:`float`)
            The effective mass in units of electron rest mass, :math:`m_0`.

        'energies' (:obj:`numpy.ndarray`)
            Band eigenvalues in eV.

        'distances' (:obj:`numpy.ndarray`)
            Distances of the k-points in reciprocal space.

        'band_id' (:obj:`int`)
            The index of the band,

        'spin' (:obj:`~pymatgen.electronic_structure.core.Spin`)
            The spin channel

        'start_kpoint' (:obj:`int`)
            The index of the k-point at which the band extrema occurs

        'end_kpoint' (:obj:`int`)
    """
    if not filenames:
        filenames = find_vasprun_files()
    elif isinstance(filenames, str):
        filenames = [filenames]

    bandstructures = []
    for vr_file in filenames:
        vr = BSVasprun(vr_file, parse_projected_eigen=False)
        bs = vr.get_band_structure(line_mode=True)
        bandstructures.append(bs)
    bs = get_reconstructed_band_structure(bandstructures, force_kpath_branches=False)

    if bs.is_metal():
        logging.error("ERROR: System is metallic!")
        sys.exit()

    _log_band_gap_information(bs)

    vbm_data = bs.get_vbm()
    cbm_data = bs.get_cbm()

    logging.info("\nValence band maximum:")
    _log_band_edge_information(bs, vbm_data)

    logging.info("\nConduction band minimum:")
    _log_band_edge_information(bs, cbm_data)

    if parabolic:
        logging.info("\nUsing parabolic fitting of the band edges")
    else:
        logging.info("\nUsing nonparabolic fitting of the band edges")

    if temperature:
        logging.error("ERROR: This feature is not yet supported!")

    else:
        # Work out where the hole and electron band edges are.
        # Fortunately, pymatgen does this for us. Points at which to calculate
        # the effective mass are identified as a tuple of:
        # (spin, band_index, kpoint_index)
        hole_extrema = []
        for spin, bands in vbm_data["band_index"].items():
            hole_extrema.extend(
                [
                    (spin, band, kpoint)
                    for band in bands
                    for kpoint in vbm_data["kpoint_index"]
                ]
            )

        elec_extrema = []
        for spin, bands in cbm_data["band_index"].items():
            elec_extrema.extend(
                [
                    (spin, band, kpoint)
                    for band in bands
                    for kpoint in cbm_data["kpoint_index"]
                ]
            )

        # extract the data we need for fitting from the band structure
        hole_data = []
        for extrema in hole_extrema:
            hole_data.extend(
                get_fitting_data(bs, *extrema, num_sample_points=num_sample_points)
            )

        elec_data = []
        for extrema in elec_extrema:
            elec_data.extend(
                get_fitting_data(bs, *extrema, num_sample_points=num_sample_points)
            )

    # calculate the effective masses and log the information
    logging.info("\nHole effective masses:")
    for data in hole_data:
        eff_mass = fit_effective_mass(
            data["distances"], data["energies"], parabolic=parabolic
        )
        data["effective_mass"] = eff_mass
        _log_effective_mass_data(data, bs.is_spin_polarized, mass_type="m_h")

    logging.info("\nElectron effective masses:")
    for data in elec_data:
        eff_mass = fit_effective_mass(
            data["distances"], data["energies"], parabolic=parabolic
        )
        data["effective_mass"] = eff_mass
        _log_effective_mass_data(data, bs.is_spin_polarized)

    return {"hole_data": hole_data, "electron_data": elec_data}


def _log_band_gap_information(bs):
    """Log data about the direct and indirect band gaps.

    Args:
        bs (:obj:`~pymatgen.electronic_structure.bandstructure.BandStructureSymmLine`):
    """
    bg_data = bs.get_band_gap()
    if not bg_data["direct"]:
        logging.info(f"Indirect band gap: {bg_data['energy']:.3f} eV")

    direct_data = bs.get_direct_band_gap_dict()
    if bs.is_spin_polarized:
        direct_bg = min(spin_data["value"] for spin_data in direct_data.values())
        logging.info(f"Direct band gap: {direct_bg:.3f} eV")

        for spin, spin_data in direct_data.items():
            direct_kindex = spin_data["kpoint_index"]
            direct_kpoint = bs.kpoints[direct_kindex].frac_coords
            direct_kpoint = kpt_str.format(k=direct_kpoint)
            eq_kpoints = bs.get_equivalent_kpoints(direct_kindex)
            k_indices = ", ".join(map(str, eq_kpoints))

            # add 1 to band indices to be consistent with VASP band numbers.
            b_indices = ", ".join([str(i + 1) for i in spin_data["band_indices"]])

            logging.info(f"  {spin.name.capitalize()}:")
            logging.info(f"    k-point: {direct_kpoint}")
            logging.info(f"    k-point indices: {k_indices}")
            logging.info(f"    Band indices: {b_indices}")

    else:
        direct_bg = direct_data[Spin.up]["value"]
        logging.info(f"Direct band gap: {direct_bg:.3f} eV")

        direct_kindex = direct_data[Spin.up]["kpoint_index"]
        direct_kpoint = kpt_str.format(k=bs.kpoints[direct_kindex].frac_coords)
        k_indices = ", ".join(map(str, bs.get_equivalent_kpoints(direct_kindex)))
        b_indices = ", ".join(
            [str(i + 1) for i in direct_data[Spin.up]["band_indices"]]
        )

        logging.info(f"  k-point: {direct_kpoint}")
        logging.info(f"  k-point indices: {k_indices}")
        logging.info(f"  Band indices: {b_indices}")


def _log_band_edge_information(bs, edge_data):
    """Log data about the valence band maximum or conduction band minimum.

    Args:
        bs (:obj:`~pymatgen.electronic_structure.bandstructure.BandStructureSymmLine`):
            The band structure.
        edge_data (dict): The :obj:`dict` from ``bs.get_vbm()`` or
            ``bs.get_cbm()``
    """
    if bs.is_spin_polarized:
        spins = edge_data["band_index"].keys()
        b_indices = [
            ", ".join([str(i + 1) for i in edge_data["band_index"][spin]])
            + f"({spin.name.capitalize()})"
            for spin in spins
        ]
        b_indices = ", ".join(b_indices)
    else:
        b_indices = ", ".join([str(i + 1) for i in edge_data["band_index"][Spin.up]])

    kpoint = edge_data["kpoint"]
    kpoint_str = kpt_str.format(k=kpoint.frac_coords)
    k_indices = ", ".join(map(str, edge_data["kpoint_index"]))

    if kpoint.label:
        k_loc = kpoint.label
    else:
        branch = bs.get_branch(edge_data["kpoint_index"][0])[0]
        k_loc = f"between {branch['name']}"

    logging.info(f"  Energy: {edge_data['energy']:.3f} eV")
    logging.info(f"  k-point: {kpoint_str}")
    logging.info(f"  k-point location: {k_loc}")
    logging.info(f"  k-point indices: {k_indices}")
    logging.info(f"  Band indices: {b_indices}")


def _log_effective_mass_data(data, is_spin_polarized, mass_type="m_e"):
    """Log data about the effective masses and their directions.

    Args:
        data (dict): The effective mass data. Formatted as a :obj:`dict` with
            the keys:

            'effective_mass' (:obj:`float`)
                The effective mass in units of electron rest mass, :math:`m_0`.

            'energies' (:obj:`numpy.ndarray`)
                Band eigenvalues in eV.

            'band_id' (:obj:`int`)
                The index of the band,

            'spin' (:obj:`~pymatgen.electronic_structure.core.Spin`)
                The spin channel

            'start_kpoint' (:obj:`int`)
                The index of the k-point at which the band extrema occurs

            'end_kpoint' (:obj:`int`)
                The k-point towards which the data has been sampled.

        is_spin_polarized (bool): Whether the system is spin polarized.
    """
    s = f" ({data['spin'].name})" if is_spin_polarized else ""

    # add 1 to band id to be consistent with VASP
    band_str = f"band {data['band_id'] + 1}{s}"

    start_kpoint = data["start_kpoint"]
    end_kpoint = data["end_kpoint"]
    eff_mass = data["effective_mass"]

    kpoint_str = kpt_str.format(k=start_kpoint.frac_coords)
    if start_kpoint.label:
        kpoint_str += f" ({start_kpoint.label})"
    kpoint_str += " -> "
    kpoint_str += kpt_str.format(k=end_kpoint.frac_coords)
    if end_kpoint.label:
        kpoint_str += f" ({end_kpoint.label})"

    logging.info(f"  {mass_type}: {eff_mass:.3f} | {band_str} | {kpoint_str}")


def _get_parser():
    parser = argparse.ArgumentParser(
        description="""
    bandstats provides information on the band gap and effective
    masses of semiconductors.""",
        epilog=f"""
    Author: {__author__}
    Version: {__version__}
    Last updated: {__date__}""",
    )

    parser.add_argument(
        "-f",
        "--filenames",
        default=None,
        nargs="+",
        metavar="F",
        help="one or more vasprun.xml files to plot",
    )
    parser.add_argument(
        "-n",
        "--nonparabolic",
        default=True,
        action="store_false",
        help=("use a nonparabolic model to fit the effective masses"),
    )
    parser.add_argument(
        "-s",
        "--sample-points",
        default=3,
        type=int,
        dest="sample_points",
        metavar="N",
        help="number of k-points to sample in fitting",
    )
    return parser


def main():
    args = _get_parser().parse_args()
    logging.basicConfig(
        filename="sumo-bandstats.log",
        level=logging.INFO,
        filemode="w",
        format="%(message)s",
    )
    console = logging.StreamHandler()
    logging.info(" ".join(sys.argv[:]))
    logging.getLogger("").addHandler(console)

    bandstats(
        filenames=args.filenames,
        num_sample_points=args.sample_points,
        parabolic=args.nonparabolic,
    )


if __name__ == "__main__":
    main()
